#define EVENT_ALLOC_BLOCK 50000  /*  this is the size of a block that gets allocated to the Events array */
#define BREAK_ALLOC_BLOCK 50000  /*  size of allocation block for AllBreaks and VisiBreaks arrays */
#define MAX_NUM_TIME_POINTS 200
#define NUM_POP_CURVE_POINTS 500

#define CYC_EVENTS(i,A)   for(i=0;i<A->NumEvents;i++)  {
#define END1CYCLE }


#ifdef UN_EXTERN
	#define GLOB 
#else
	 #define GLOB extern
#endif

#include "ECA_intervals.h"

/* here is an enum list for the types of nodes */
enum NodeType 
{
	Unset,	/* 0 is for a node that has not been set to a type yet */
	Leaf,  /* 1 is for the ends of the tree */
	Recomb, /* 2 is for a recombination node */
	Coalesce  /* 3 is for a coalescent event  */
};

enum BranchNums
{
	Left,   /* the left branch is 0 */
	Right,  /* the right branch is 1 */
	Middle,  /* the middle branch is 2 */
	UnsetB  /* if you haven't set it, give it a three  */
};


typedef struct
{
	struct node_struct *top;    /*  points to the node that is above (later in time) the branch */
	struct node_struct *bottom; /*  points to the node that is below (earlier in time) the branch */
	
	double lo;  /*  if the position on the chromosome is greater than lo and less than high, then */
	double hi;	/*  this branch exists.   */
	
	double wt;  /*  "weight"---just max(0,hi-lo)  ---> the length of interval within which a recombination */
				/*  point will affect the genealogy. */
	
	enum BranchNums TS;  /*  the side into which the top of the branch enters (left, middle, right) */
	enum BranchNums BS;	/*  the side from which the bottom of the branch exits (going up). */
	
	int *NumThrough;   /*  The number of paths traced from all the leaves through the branch given a position on the segment.   */
						/*  NumThrough[0] will be for a position on an even numbered segment, NumThrough[1] will be for  */
						/*  a position on an odd-numbered segment (i.e. these segments are the bits between all the */
						/*  recombination points).  And NumThrough[2] will be the sum of NumThrough[0] and NumThrough[1] */
						/*  after running paths through branches from all the leaves.  If this is zero after checking all */
						/*  the segments, then that branch is Dead on the whole interval we are looking at over all the segments */
	
	int NumLeaves;
	
	IntervalsStruct *AncBits;  /*  Ancestral bits:  holds the intervals that correspond to material */
							   /*  ancestral to at least one of the sampled sequences */
							   
	IntervalsStruct **AncBitsBySeg;  /*  an array of Intervals of ancestral material, one each for */
									 /*  each of the leaves */
	IntervalsStruct *OverallAncBits; /*  to store the intersection of all the AncBitsBySegs */
	
} branch_struct;



typedef struct node_struct
{
	enum NodeType type;
	
	int ID;  /*  Just an ID number for it if I care to have one */
	
	int BeenVisited;  /*  particular important for recombination nodes, used in post-order */
					  /*  traversal.  Tells whether a node has been visited previously or not. */
					  /*  0 = NotVisited, then any positive number can indicate the number of visits */
					  
	int NumThrough;		/*  when determining the MRCA, *for a single position* we trace paths back from each of  */
						/*  the leaves and count how many of those paths go through a node.  The first coalescent */
						/*  node, going back in time, that has NumLeaves paths go through it is the MRCA. */
	
	int VisitOrder;
	double time;  /*  the time when this node occurs */
	
	double x;  /*  the position of the recombination if this is a recomb event */
	
	branch_struct **Br;  /*  the branches which will be Left, Right, and Middle */
						 /*  i.e. if this is a coalescent node with the Middle branch coming */
						 /*  out the top, then Br[Middle]->top points up to the next node and */
						 /*  Br[Middle]->bottom points back at this node itself  */
						 
	double *Pos;	/*  these are for 2-D and 3-D drawing of the trees. */
					/*  Pos[0] is the X-coordinate, Pos[1] is the Y-coordinate, and Pos[2] */
					/*  is the Z-coordinate */
	
	double **P;  /*  this is an array of Pos's.  So we have one array for each of */
				 /*  of the Middle, Left, and Right branches.  Better for drawing rectilinear trees. */


} node_struct;




/*  Global variable stuff to be able to de-allocate quickly  */
typedef struct
{
	node_struct *PtNS_Array[EVENT_ALLOC_BLOCK];
	branch_struct *PtBS_Array[3 * EVENT_ALLOC_BLOCK];
	
	int NS_ctr;
	int BS_ctr;
} Mems_struct;




typedef struct node_list_tag  /*  for a straight linked list of nodes going back in time */
{
	node_struct *Data;
	struct node_list_tag *Next;
	
} node_list;


typedef struct branch_list_tag
{
	branch_struct *Data;
	struct branch_list_tag *Next;
	
} branch_list;



/*  this structure is designed to hold information about the growth rate of populations, */
/*  and such, that the user has put into a file.  the info in this struct gets processed */
/*  in order to set the parameters in the popgrowth_struct itself. */
typedef struct
{
	int IsTimePoint;
	int IsSizePoint;
	
	double Time;
	int TimeSpecified;
	
	double EntrySize;
	int EntrySizeSpecified;
	
	double ExitSize;
	int ExitSizeSpecified;
	
	double Rate;
	int RateSpecified;
	
} popgrowth_element_struct;

/*  this structure is designed to hold information about the growth rate of populations */
/*  and such that the user has put into a file.  the info in this struct gets processed */
/*  in order to set the parameters in the popgrowth_struct itself. */
typedef struct
{
	int NZero;  /*  the population size at present day.  This  */
				/*  will be identical to P[0]->ExitSize */
	int NumPoints;
	
	popgrowth_element_struct *P[MAX_NUM_TIME_POINTS];
	
} popgrowth_input_struct;


/*  this structure holds the info about N^+,  t, \nu, and the sum of the \nu's */
/*  for actually doing the transformation between scaled and unscaled time */
typedef struct
{
	double t;		/*  the time of each "stage" in unscaled time */
	double Nplus;  /*  does not need to be an int!!  We're in continuous asymptopia anyway.   */
	double r;		/*  the exponential rate of decline or increase */
	double nu;		/*  see the TeX document about this... */
	double sum_nu;

} popgrowth_struct;

/*  this structure is designed to hold information about the growth rate of populations */
/*  the population history can be divided into a series of "pieces" going back in time */
/*  as in "piece-wise" exponential growth.  Each series  */
typedef struct
{

	int UsePopGrowth;  /*  set to zero if you want stable populatoin size, non-zero if you  */
						/*  want to use what is in the struct */
	int Num;  /*  the number of stages */
	popgrowth_struct *P;
	
	/*  and now we want to keep a list of popsizes at fairly regular intervals */
	/*  through time.  This is for plotting the pop-size curve.  */
	double ts[NUM_POP_CURVE_POINTS];
	double ns[NUM_POP_CURVE_POINTS];
	
	double MaxN;  /*  the max value that the pop size reaches */
	
} popgrowth_container;


/*  structure to hold all the parameters.  Currently that isn't too many, but it will be easy */
/*  to add pop growth parameters, etc., eventually */
typedef struct
{
	int NumLeaves;
	double rho;  /*  scaled recombination rate */
	double L;  	 /*  length of the segments under consideration */
	double Lo;	 /*  Left and right endpoints of the segment */
	double Hi;
	
	popgrowth_container *pc;
	
} parm_struct;


/*  a structure to hold all the information relevant to a single Ancestral Recombination Graph */
/*  The idea is to be able to simulate values into this struct, then hand it off to be graphed */
/*  or averaged or analyzed in whatever way necessary. */
typedef struct 
{
	parm_struct *Pars;  /*  pointer to the parameters governing the properties of this graph */
	
	int NumEvents;  /*  The number of coalescences and recombinations in the AR_graph */
	
	node_struct **Events;  /*  an array of the events, time-ordered, subscripted from 0 to NumEvents-1 */
	node_struct **Leaves;  /*  an array of pointers to the leaves of the graph */
	
	node_struct *Root;  /*  pointer to the root of the tree */
	
	branch_list *BrList;  /*  Temporary list that gets used during simulation. */
	branch_list **BrArray;  /*  an array of pointers to the branches active just below event [i] but after */
							/*  event [i-1].  For example, it would include the two branches that will coalesce */
							/*  at event i. */
	
	int NumAllBreaks;	/*  the number of breakpoints occurring on any segment in the AR_graph */
	double *AllBreaks;  /*  an array of the positions of all the breakpoints that occurred anywhere */
						/*  in the graph */
					
	int NumVisiBreaks;  	/*  the number of breakpoints that actually cause differences in the ancestry */
						/*  of the sampled segments---the are VISIble.   */
	double *VisiBreaks;	/*  array of the positions of the "visible" breaks.   */
	int *MRCAofVisiBreaks;  /* array of indexes of the most recent common ancestor for  points 
									 just to the right of the i-th visibreak.  i.e. Events[MRCAofVisiBreaks[i]] is the 
									 common ancestral node of the positions right of the recombination point
									 indexed by i and left of one indexed by i+1 */
	double LeftTimeToMRCA;  /*  Time to the MRCA from the left endpoint of the segment  */
	int LeftMRCA;			/*  Index of the event which is the MRCA of the left endpoint */
	
	IntervalsStruct **AllAncBits;  /*  An array, one for each event, of the union of the  */
								   /*  OverallAncBits associated with each of the branches */
								   /*  currently active above the last branch. */
/* 	double BreaksInAncBits[500]; */
/* 	int NumBreaksInAncBits; */

} AR_graph_struct;






/*  GLOBAL VARIABLES */
GLOB Mems_struct gMem;






/*  FUNCTION PROTOTYPES */

/******************************************************************************************/
/*                                                                                        */
/*		              LOW-LEVEL MEMORY ALLOCATION/DEALLOCATION                            */
/*                                                                                        */  
/******************************************************************************************/
node_struct *AllocNode(int inID);
branch_struct *AllocBranch(parm_struct *Pars);
void DeallocNode(node_struct *N);
void DeAllocBranch(branch_struct *B);




/******************************************************************************************/
/*                                                                                        */
/*		                PRINTING AND SIMPLE OUTPUT FUNCTIONS                              */
/*                                                                                        */  
/******************************************************************************************/
void PrintBranch(branch_struct *B);
void PrintBranchList(branch_list *B);
void PrintNode(node_struct *N);
const char *NodeType2str(enum NodeType node);
const char *BranchNum2str(enum BranchNums nums);




/******************************************************************************************/
/*                                                                                        */
/*		                     GENEALOGY SIMULATION FUNCTIONS                               */
/*                                -- Nuts and Bolts --                                    */  
/******************************************************************************************/
branch_list *AppendBranchToList(branch_list *B, branch_struct *StickItInThere);
IntervalsStruct *UnionOverBranches(branch_list *B);
IntervalsStruct *RecombIntersect(IntervalsStruct *InAnc, double x, 
				enum  BranchNums RightLeft, parm_struct *Pars);
IntervalsStruct *CreateLeafInterval(int OnOff, parm_struct *Pars);
node_struct *CreateLeaf(int ID, parm_struct *Pars);
node_struct *CoalesceTwoRandomBranches(branch_list **Bl, double Time, 
				int *NumBranches, int ID,parm_struct *Pars);
node_struct *SplitOneRandomBranch(branch_list **Bl, double Time, int *NumBranches,
			 	int ID, double *Position,parm_struct *Pars);
node_struct *CreateCoalescence(branch_struct *A, branch_struct *B, 
				double t, int ID, parm_struct *Pars);
node_struct *CreateRecomb(branch_struct *A, double x, double t, int ID, parm_struct *Pars);





/******************************************************************************************/
/*                                                                                        */
/*		                   GENEALOGY SIMULATION/DESTRUCTION                               */
/*                             -- Big Ol' Functions --                                    */  
/******************************************************************************************/
AR_graph_struct *NewAR_graph(parm_struct *P);
void ObliterateNodesAndBranches(void);
void Destroy_AR_graph(AR_graph_struct *A);




/******************************************************************************************/
/*                                                                                        */
/*		          ANALYZING GENEALOGY AFTER IT'S BEEN SIMULATED                            */
/*                                                                                        */  
/******************************************************************************************/
int FinalCoalesce(AR_graph_struct *A, double Lo, double Hi);
int BranchIsActive(branch_struct *B, double Lo, double Hi);
void SetNumThroughsToZero(AR_graph_struct *AR, int ZeroOneTwo);
int GenealogyHasChanged(AR_graph_struct *A, int MRCA1, int MRCA2);
int FindMRCAbyPaths(AR_graph_struct *A, double x, int OneOrZero);
void ClimbTree(node_struct **N, double TopTime, double Pos, int OneOrZero);
int CompDoubles (const void *a, const void *b);
void FindVisiBreaks(AR_graph_struct *A);



/******************************************************************************************/
/*                                                                                        */
/*          FUNCTIONS TO PREPARE FOR PLOTTING THE GENEALOGY GRAPHICALLY                   */
/*                                                                                        */  
/******************************************************************************************/
void MapNodesGivenLeaves(AR_graph_struct *A);
void PostOrderMapWPos(node_struct *Node, int *LV, double T, int LeafNum, int *NV, double Pos);
void MapNodesWPos(AR_graph_struct *A, double Pos);



/******************************************************************************************/
/*                                                                                        */
/*          FUNCTIONS IN THE FILE HapTypeGrowthPars.c				                      */
/*                                                                                        */  
/******************************************************************************************/
popgrowth_element_struct *AllocPES(void);
popgrowth_input_struct *ReadPopGrowthFile(char *FileName);
popgrowth_container *ProcessPopGrowth(popgrowth_input_struct *PIP);
void PrintPopGrowth(popgrowth_container *pc);
double T_From_Th(popgrowth_container *pc, double th, double Units);
double Th_From_T(popgrowth_container *pc, double t, double Units);



















