/*  Header file for the analysis data structures and function prototypes  */

#include "MCTypesEtc.h"

/*  Here is a data structure for particular points on the interval */
typedef struct 
{
	double Pos;   /*  the position of the point */
	
	dval *LeftBreakDist;    /*  to record the distances to the left and right breaks */
	dval *RightBreakDist;
	
	dval *SegLength;  /*  to record the total length of the segment that the point is in */
	
	dval *TimeToMRCA;  /*  the time to the MRCA for this location. */
	
	
} pts_anal_struct;


typedef struct
{
	double Xk;  /*  the length to the first recombination as a function of k (in the TPB paper) */
	double PrHi_less_H2;  /*  probability that the time to MRCA for the second segment is greater than */
						  /*  the time to MRCA of the first position. in TPB */
	double E_Num_VBs;  /*  expected number of VisiBreaks.  Hudson and Kaplan's result. */
	double E_NumShifts;  /*  Griffiths and Marjoram's result for S_k(rho).  Expected number of */
						 /*  shifts between MRCA's */
	double E_Dist_b_t_VBs;  /*  1/ E_Num_VBs */
	double E_Dist_b_t_Shifts;  /*  1/E_NumShifts */
	
} expected_values_struct;

typedef struct
{
	int ResetTheoretical;  /*  1 if the theoretical dsn needs to be reset (if rho changed...) */
	
	char **Names;
	int NumNames;
	char TheoreticalDsnString[100];
	char SimulatedDsnString[100];
	char TheoreticalExpString[100];
	char SimulatedExpString[100];
	
	double Xks[500];   /*  500 should be more than enough */
	double Pdf[500];
	
	int NumPoints;  /*  the number of points to evaluate the density at */
	double MaxPoint; /*  the highest value of Xk one will record */
	double MaxDensity;  /*  highest peak of the density */
	double step;	 /*  the step size to use:  MaxPoint/(NumPoints-1) 'cuz we always start at zero,	 */
					/*  even though we will not try to graph that point. */
	

}  Xk_dist_struct;


/*  Here is a data structure to keep track of some things */
typedef struct
{
	parm_struct *Pars;	/*  For the parameters governing this simulation */
	
	expected_values_struct *E;

	int NumPoints;      /*  The number of specific locations that one is collecting data from */
	pts_anal_struct **Points;  /*  Array of the specific locations with space to store info */
								/*  about the lengths on either side of them */
								
	DvalCovarSquare *Covars;  /*  to record the covariances in time to MRCA between all the points */
					/*  this will be NumPoints by NumPoints.  Does not need a histogram. */
										
	dval *SegLengths;  /*  a single struct to record the overall distribution on segment lengths */
					   /*  this one will need a histogram */
					   
	int NumLengthBins; /*  the number of bins of different lengths to have for the  */
					   /*  TimesToMRCA_BySegLengths array.   */
					   
	Xk_dist_struct *Xk_dist;
					   
	dval **TimesToMRCA_BySegLengths;  
	
	ival *NumVisiBreaks;
	ival *NumInvisiBreaks;
	ival *NumAllBreaks;
	
	dval *LengthToFirstRecomb;
	dval *ProbTree2Higher;
	
	ival *ShiftsBetweenMRCAs;
	

} haptype_anal_struct;



/*  Prototypes: */
pts_anal_struct *AllocPtsAnal(parm_struct *P, double Pos);
haptype_anal_struct *AllocHapTypeAnal(parm_struct *P, int NumPoints, double *Pos);
void InitPointsToZero(pts_anal_struct *P);
void InitHapTypeAnalToZero(haptype_anal_struct *H);
int RecursiveFindSeg(double P, double *A, int lo, int hi);
void TransferToPtsAnal(pts_anal_struct *P, AR_graph_struct *A);
void AccumulateAR_graph(AR_graph_struct *A, haptype_anal_struct *H );
void PrintAnalysis(haptype_anal_struct *H, int NumTreesSimulated);
expected_values_struct *ComputeExpectedValues(parm_struct *P, expected_values_struct *temp );
void ComputeXkDist(Xk_dist_struct *X, parm_struct *P, dval **LTFR);











