
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "MathStatRand.h"
#include "HapType.h"
#include "ranlib.h"
#include "HapTypeAnal.h"



pts_anal_struct *AllocPtsAnal(parm_struct *P, double Pos)
{	
	pts_anal_struct *temp;
	
	temp = (pts_anal_struct *)malloc(sizeof(pts_anal_struct));
	
	temp->Pos = Pos;
	
	temp->LeftBreakDist = AllocDval(P->Lo,P->Hi,(P->Hi - P->Lo) / 100.0);
	temp->RightBreakDist = AllocDval(P->Lo,P->Hi,(P->Hi - P->Lo) / 100.0);
	temp->SegLength = AllocDval(P->Lo,P->Hi,(P->Hi - P->Lo) / 100.0);
	temp->TimeToMRCA = AllocDval(0.0,0.0,0.0);
	
	return(temp);
}


haptype_anal_struct *AllocHapTypeAnal(parm_struct *P, int NumPoints, double *Pos)
{
	
	int i;
	haptype_anal_struct *temp;
	
	temp = (haptype_anal_struct *)malloc(sizeof(haptype_anal_struct));
	
	temp->NumPoints = NumPoints;
	
	/*  copy the parameters over */
	temp->Pars = (parm_struct *)malloc(sizeof(parm_struct));
	
	temp->Pars->rho = P->rho;
	temp->Pars->Lo = P->Lo;
	temp->Pars->Hi = P->Hi;
	temp->Pars->L = temp->Pars->Hi - temp->Pars->Lo;
	temp->Pars->NumLeaves = P->NumLeaves;
	
	temp->E = (expected_values_struct *)malloc(sizeof(expected_values_struct));
	 ComputeExpectedValues(P,temp->E);
	
	temp->Points = (pts_anal_struct **)calloc(temp->NumPoints, sizeof(pts_anal_struct *));
	for(i=0;i<temp->NumPoints;i++)  {
		temp->Points[i] = AllocPtsAnal(P,Pos[i]);
	}
	
	temp->Covars = AllocDvalCovarSquare(temp->NumPoints);
	
	
	temp->SegLengths = AllocDval(P->Lo,P->Hi,(P->Hi - P->Lo) / 100.0);
	
	temp->NumLengthBins = 0;
	
	
	
	
	
	temp->TimesToMRCA_BySegLengths = DvalVector(0,100,0.0,0.0,0.0);
	
	temp->NumVisiBreaks = AllocIval(0.0,0.0,0.0);
	temp->NumInvisiBreaks = AllocIval(0.0,0.0,0.0);
	temp->NumAllBreaks = AllocIval(0.0,0.0,0.0);
	temp->ShiftsBetweenMRCAs = AllocIval(0.0,0.0,0.0);
	
	
	temp->ProbTree2Higher = AllocDval(0.0,0.0,0.0);
	
	
	/*  length to first recomb stuff */
	temp->Xk_dist = (Xk_dist_struct *)malloc(sizeof(Xk_dist_struct));
	
	ComputeXkDist(temp->Xk_dist,P, &(temp->LengthToFirstRecomb));

	
	
	
	/*  deal with  */
	
	return(temp);
}


/*  set Points stuff to zero */
void InitPointsToZero(pts_anal_struct *P)
{
	InitDvalToZero(P->LeftBreakDist);
	InitDvalToZero(P->RightBreakDist);
	InitDvalToZero(P->SegLength);
	InitDvalToZero(P->TimeToMRCA);
	
}


/*  set everything to zero to get ready to start accumulating values */
void InitHapTypeAnalToZero(haptype_anal_struct *H)
{
	int i;
	
	for(i=0;i<H->NumPoints;i++)  {
		 InitPointsToZero(H->Points[i]);
	}
	
	InitDvalCovarSquareToZero(H->Covars);
	
	InitDvalToZero(H->SegLengths);
	
	for(i=0;i<H->NumLengthBins;i++)  {
		InitDvalToZero(H->TimesToMRCA_BySegLengths[i]);
	}
	
	InitIvalToZero(H->NumVisiBreaks);
	InitIvalToZero(H->NumInvisiBreaks);
	InitIvalToZero(H->NumAllBreaks);
	
	InitDvalToZero(H->LengthToFirstRecomb);
	InitDvalToZero(H->ProbTree2Higher);
	
	InitIvalToZero(H->ShiftsBetweenMRCAs);
}


/* 
	Given a sorted array of doubles, A, this
	returns the index of the value that is to the left of the 
	point P.  Returns -1 if P is less than A[lo]
	
	It should be called with Lo and Hi being the min and
	the max index of the array A.
	
	This is based on binary search.
	
	Only call this if P > A[0].  
*/
int RecursiveFindSeg(double P, double *A, int lo, int hi)
{	
	int mid = (lo + hi)/2;
	
	/*  sort of silly thing here to catch.  SHould make the more intelligently in */
	/*  the future, but sometimes there are no breakpoints in A so A is null and */
	/*  we don't want to try to look inside it.  And of course, we know that all */
	/*  points will have the same ancestor  as the LeftMost at that point.  We know */
	/*  this is happening because lo == 0 and hi == -1 */
	
	if(lo ==0  && hi == -1)  {
		return(-1);  /*  this lets it know it goes to the leftmost. */
	}
	
	
	if(P<A[lo])
		return(-1);
	else if(P>A[hi])
		return(hi);
	if(lo>=hi)  {
		printf("\nlo >= hi in FindSegIndex().  Quitting!...\n\n)");
		exit(1);
	}
	if(hi == lo + 1)  /*  if we've whittled it down to a single interval */
		return(lo);
	
	else if(P<A[mid])  
		return(RecursiveFindSeg(P,A,lo,mid));
	else if(P>A[mid])
		return(RecursiveFindSeg(P,A,mid,hi));
	else if(P==A[mid])
		return(mid);
	
	/*  returns a -2 if it got here.  This is a problem!!! It should never get here!! */
	return(-2);
}


/*  Given a pts_anal_struct and an AR_graph_Struct */
/*  store the values to the pts_anal_struct and increment it */
void TransferToPtsAnal(pts_anal_struct *P, AR_graph_struct *A)
{
	int temp;
	/*  find the left break distance and the TimeToMRCA */
	temp = RecursiveFindSeg(P->Pos,A->VisiBreaks,0,A->NumVisiBreaks-1);
	
	if(temp < 0)  {
		if(temp==-1) {
			P->LeftBreakDist->v = P->Pos;
			P->TimeToMRCA->v = A->LeftTimeToMRCA;
		}
		else {
			printf("RecursiveFindSeg() returned < -1.  Bad News Dude!");
			printf("\nExiting....\n\n");
			exit(1);
		}
	}
	else {
		P->LeftBreakDist->v = P->Pos - A->VisiBreaks[temp];
		P->TimeToMRCA->v = A->Events[A->MRCAofVisiBreaks[temp]]->time;
	}
	
	/*  find the right break distance */
	if(temp==A->NumVisiBreaks-1)
		P->RightBreakDist->v = A->Pars->Hi - P->Pos;
	else
		P->RightBreakDist->v = A->VisiBreaks[temp+1] - P->Pos;
		
	/*  now the SegLength is easy to compute */
	P->SegLength->v = P->LeftBreakDist->v + P->RightBreakDist->v;

	/*  And now we increment all those goodies */
	IncrementDval(P->LeftBreakDist);
	IncrementDval(P->RightBreakDist);
	IncrementDval(P->SegLength);
	IncrementDval(P->TimeToMRCA);	
}


/*  Gather all the information about the current AR_graph struct and */
/*  store it in the MC_Types, and increment them while we are at it. */
void AccumulateAR_graph(AR_graph_struct *A, haptype_anal_struct *H )
{
	int i, currentMRCA, shifts;
	dval **array;  /*  temporary */
	
	array = (dval **)calloc(H->NumPoints, sizeof(dval *));
	
	/*  get the data for the points and increment them too. */
	for(i=0;i<H->NumPoints;i++)  {
		TransferToPtsAnal(H->Points[i],A);  /*  this function also increments them */
		
		/* point the elements of array to where they need to go */
		array[i] = H->Points[i]->TimeToMRCA;
	}
	
	/*  then deal with the covar square */
	IncrementDvalCovarSquare(H->Covars,array);
	
	
	/*  then do the other things: */
	/*  NumVisiBreaks: */
	H->NumVisiBreaks->v = A->NumVisiBreaks;
	IncrementIval(H->NumVisiBreaks);
	
	/*  Length to First Recomb */
	if(A->NumVisiBreaks >= 1) {
		H->LengthToFirstRecomb->v = A->VisiBreaks[0];
		IncrementDval(H->LengthToFirstRecomb);
	}
	
	
	/*  ProbTree2Higher.  Note that rather than comparing times */
	/*  which are doubles, and therefore may have some problems with */
	/*  comparisons (< > or =) just use the ID of the events of the */
	/*  different MRCA's */
	if(A->NumVisiBreaks >= 1)  {
		if(A->Events[A->MRCAofVisiBreaks[0]]->ID > A->Events[A->LeftMRCA]->ID)
			H->ProbTree2Higher->v = 1.0;
		else 
			H->ProbTree2Higher->v  = 0.0;
		IncrementDval(H->ProbTree2Higher);
	}
	else {
		H->ProbTree2Higher->v = 1.0;
		IncrementDval(H->ProbTree2Higher);
	}
	

	/*  gotta count the number of shifts between MRCA's: */
	/*  do this by cycling through the visibreaks and seeing how many shifts */
	/*  there are in MRCAofVisiBreaks */
	currentMRCA = A->LeftMRCA;
	shifts = 0;
	for(i=0;i<A->NumVisiBreaks;i++)  {
		if(A->MRCAofVisiBreaks[i] != currentMRCA)
			shifts++;
			
		 currentMRCA = A->MRCAofVisiBreaks[i];
	}
	
	H->ShiftsBetweenMRCAs->v = shifts;
	IncrementIval(H->ShiftsBetweenMRCAs);




	free(array);

}


void PrintAnalysis(haptype_anal_struct *H, int NumTreesSimulated)
{
	int i,j;
	char tempStr[300];
	FILE *out;
	double z,p,result;
	
	/*  Set the name for the output file */
	sprintf(tempStr,"Rho%.1fLe%dLo%.1fHi%.1f",H->Pars->rho,H->Pars->NumLeaves,
				H->Pars->Lo,H->Pars->Hi);
	
	out = fopen(tempStr,"w");
	
	fprintf(out,"\nLo = %.4f\nHi = %.4f\nrho = %.4f\nNumLeaves = %d\n",
						H->Pars->Lo,H->Pars->Hi,H->Pars->rho,H->Pars->NumLeaves);
	fprintf(out,"Number of Trees Simulated = %d\n\n", NumTreesSimulated);
	
	/*  print the column titles first */
	fprintf(out,"RHO = %f  NUMREPS = %d",H->Pars->rho,H->Points[0]->TimeToMRCA->NumAved);
	fprintf(out,"\nPosition: ");
	for(i=0;i<H->NumPoints;i++) {
		fprintf(out,"  %.6f",H->Points[i]->Pos);
	}
	
	/*  then printf out the table */
	for(i=0;i<H->NumPoints;i++) {
		fprintf(out,"\n%.6f",H->Points[i]->Pos);
		for(j=0;j<H->NumPoints;j++)  {
			fprintf(out,"   %f",H->Covars->Sq[i][j]->Var);
		}
	}
	
	/*  Now print out roughly the same thing, but use the analytical result */
	fprintf(out,"\n\n\nAnalyticalResult:  ***WARNING!  ONLY VALID FOR NUM_LEAVES = 2 ****\n");
	fprintf(out,"Position: ");
	for(i=0;i<H->NumPoints;i++) {
		fprintf(out,"  %.6f",H->Points[i]->Pos);
	}
	
	/*  then printf out the table */
	for(i=0;i<H->NumPoints;i++) {
		fprintf(out,"\n%.6f",H->Points[i]->Pos);
		for(j=0;j<H->NumPoints;j++)  {
			z = fabs(H->Points[i]->Pos - H->Points[j]->Pos);
			p = H->Pars->rho;
			result = (p*z + 18.0)/ (p*p*z*z + 13.0 *p*z + 18.0);
			fprintf(out,"   %f",result);
		}
	}
	
	/*  then print out the other expected values and such */
	fprintf(out,"\n\n\n\n");
	fprintf(out,"\nLength to first recombination:");
	fprintf(out,"\n\tObserved = %.6f   Expected = %.6f    Number of Observations = %d",
				H->LengthToFirstRecomb->Ave,   H->E->Xk, H->LengthToFirstRecomb->NumAved);
	
	fprintf(out,"\n\nProb that second tree is higher than first:");
	fprintf(out,"\n\tObserved = %.6f   Expected = %.6f    Number of Observations = %d",
				H->ProbTree2Higher->Ave,   H->E->PrHi_less_H2, H->ProbTree2Higher->NumAved );
	
	fprintf(out,"\n\nNumber of recombinations in active ancestral material:");
	fprintf(out,"\n\tObserved = %.6f   Expected = %.6f    Number of Observations = %d",
			  H->NumVisiBreaks->Ave,    H->E->E_Num_VBs, H->NumVisiBreaks->NumAved);
	
	fprintf(out,"\n\nNumber of shifts between MRCAs:");
	fprintf(out,"\n\tObserved = %.6f   Expected = %.6f    Number of Observations = %d",
					0.0,H->E->E_NumShifts,0);
	
	fprintf(out,"\n\nDistance between recombinations in active ancestral material:");
	fprintf(out,"\n\tObserved = %.6f   Expected = %.6f   Number of Observations = %d",
					0.0,H->E->E_Dist_b_t_VBs,0);
	
	fprintf(out,"\n\nDistance between shifts between MRCAs:");
	fprintf(out,"\n\tObserved = %.6f   Expected = %.6f   Number of Observations = %d",
					0.0,H->E->E_Dist_b_t_Shifts,0);
	
	fclose(out);
	
}




expected_values_struct *ComputeExpectedValues(parm_struct *P, expected_values_struct *temp )
{	
	int i,k=P->NumLeaves;
	
	


	/*  Xk */
	if(k <= 2) 
		temp->Xk = -log(0.0);  /*  infinity */
	else {
		temp->Xk = 0.0;   /*  initialize to accumulate a sum */
		for(i=1;i<=k-2;i++)  {
			temp->Xk += pow(-1,i-1) * exp( LogFactorial(k-1) - LogFactorial(i) 
						- LogFactorial(k-i-2) ) * log(i+1);
		}
		/*  at the end we multiply by 1/rho */
		temp->Xk /= P->rho;
	}
	
	
	/*  PrHi_less_H2 */
	temp->PrHi_less_H2 = 0.0;   /*  initialize to accumulate a sum */
	for(i=1;i<=k;i++)  {
		temp->PrHi_less_H2 += pow(-1,i) * exp( LogFactorial(k-1) - LogFactorial(i-1) 
					- LogFactorial(k-i) ) * (log(i+1) + log(i));
	}
	temp->PrHi_less_H2 *= .5 * ((double)k-1.0);


	
	/*  E_Num_VBs */
	temp->E_Num_VBs = 0.0;
	for(i=1;i<=k-1;i++)  {
		temp->E_Num_VBs += 1.0/(double)i;
	}
	temp->E_Num_VBs *= P->rho * (P->Hi - P->Lo);

	
	/*  E_NumShifts */
	temp->E_NumShifts = P->rho * (1.0 - 2.0/ (double)(k*(k+1)) )  * (P->Hi - P->Lo);
	
	/*  E_Dist_b_t_VBs */
	temp->E_Dist_b_t_VBs = 1.0 / temp->E_Num_VBs;
	
	/*  E_Dist_b_t_Shifts */
	temp->E_Dist_b_t_Shifts = 1.0 / temp->E_NumShifts;
	
	return(temp);
}



/*
	computes the distribution of the length to the first recombination and stores that 
	in X.  Then at the end, it reallocates a new dval to LengthToFirstRecombination,
	passed in as an output parameter LTFR, with
	the right values in the hist_struct.  currently does not free the previously dval.  This
	is a memory leak, but I'll deal with it later.
*/
void ComputeXkDist(Xk_dist_struct *X, parm_struct *P, dval **LTFR)
{	
	int i,k=P->NumLeaves,j;
	double z,max=0.0, pdf, sum;
	int go=1; /*  start it at one to account for the zero point */
	
	X->ResetTheoretical = 0;  /*  set that for now */
	
	sprintf(X->TheoreticalDsnString,"Theor. Dsn.");
	sprintf(X->SimulatedDsnString,"Simul. Dsn.");
	sprintf(X->TheoreticalExpString,"Theor. Expctn.");
	sprintf(X->SimulatedExpString,"Simul. Expctn.");
	
	X->Names = (char **)calloc((unsigned long)4,sizeof(char *));
	X->NumNames = 2;
	X->Names[0] = X->TheoreticalDsnString;
	X->Names[1] = X->SimulatedDsnString;
	X->Names[2] = X->TheoreticalExpString;
	X->Names[3] = X->SimulatedExpString;
	
	
	/*  go through the dsn once in steps of size .01 and find the max, and then find the */
	/*  point at which the current value is 1/500 of the max.  That is where we will */
	/*  chop it off. */
	for(z=.01;go;z+=.01) {
		sum = 0.0;
		for(i=1;i<=k-1;i++)  {
			sum += pow(-1,i-1) * exp(LogFactorial(k-1) - LogFactorial(i-1) - LogFactorial(k-i-1)) *
					(1.0 /  pow(i + P->rho * z, 2.0) );
						
		}
		
		pdf = P->rho * sum;
		
		if(pdf>max)
			max = pdf;
		
		if(max/pdf > 500.0  || z > P->Hi)
			go = 0;
	}
	
	X->MaxDensity = max;
	X->MaxPoint = z;
	X->NumPoints = 100;  /*  lets just set this here like this to always be 100 */
	X->step = X->MaxPoint / (double)(X->NumPoints-1);
	
	/*  now, fill in the values: */
	for(j=1;j<X->NumPoints;j++)  {
		
		z = (double)j * X->step;
		X->Xks[j] = z;
		
		sum = 0.0;
		for(i=1;i<=k-1;i++)  {
			sum += pow(-1,i-1) * exp(LogFactorial(k-1) - LogFactorial(i-1) - LogFactorial(k-i-1)) *
					(1.0 /  pow(i + P->rho * z, 2.0) );
						
		}
		
		X->Pdf[j] = P->rho * sum;
		
	}	
	
	/*  and at the end deal with: */
	*LTFR = AllocDval(0.0,X->MaxPoint,X->step);
	
}
