
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "HapType.h"
#include "ranlib.h"


/*  note!  Prototypes for these functions are in HapType.h */

/*  some #defines for the conditionals */
#define NewSP(Q)  (strcmp(Q,"sp")==0)  || (strcmp(Q,"SizePoint")==0)  || (strcmp(Q,"SP")==0)
#define NewTP(Q)  (strcmp(Q,"tp")==0)  || (strcmp(Q,"TimePoint")==0)  || (strcmp(Q,"TP")==0)
#define NewR(Q)  (strcmp(Q,"Rate")==0)  || (strcmp(Q,"rate")==0)  || (strcmp(Q,"r")==0)  || (strcmp(Q,"R")==0)
#define NewEntry(Q)  (strcmp(Q,"Entry")==0)  || (strcmp(Q,"Ent")==0)  || (strcmp(Q,"ent")==0) || (strcmp(Q,"entry")==0)
#define NewExit(Q)  (strcmp(Q,"Exit")==0)  || (strcmp(Q,"Ex")==0)  || (strcmp(Q,"ex")==0) || (strcmp(Q,"exit")==0)
#define NewTime(Q)  (strcmp(Q,"Time")==0)  || (strcmp(Q,"time")==0)  || (strcmp(Q,"T")==0) || (strcmp(Q,"t")==0)
#define NewIPS(Q)  (strcmp(Q,"InitPopSize")==0)  || (strcmp(Q,"NZero")==0)  || (strcmp(Q,"IPS")==0) || (strcmp(Q,"nzero")==0)



popgrowth_element_struct *AllocPES(void)
{
	popgrowth_element_struct *temp;
	
	temp = (popgrowth_element_struct *)malloc(sizeof(popgrowth_element_struct));
	
	temp->IsTimePoint = 0;
	temp->IsSizePoint = 0;
	
	temp->Time = -1.0;
	temp->TimeSpecified = 0;
	
	temp->EntrySize = -1.0;
	temp->EntrySizeSpecified = 0;
	
	temp->ExitSize= -1.0;
	temp->ExitSizeSpecified = 0;
	
	temp->Rate = 0.0;
	temp->RateSpecified = 0;
	
	return(temp);
}


/*  reads a file and then returns a popgrowth input */
/*  struct pointer */
popgrowth_input_struct *ReadPopGrowthFile(char *FileName)
{
	FILE *in;
	popgrowth_input_struct *pip;
	char S[200];
	double temp;
	int scanned;
	int i;  /*  will end up being the number of time points */
	
	pip = (popgrowth_input_struct *)malloc(sizeof(popgrowth_input_struct));
	pip->NumPoints = 0;
	
	/*  open the file and then read the initial popsize */
	if( (in = fopen(FileName,"r")) == 0)  {
		printf("\n\nCan't find file %s.\n\nExiting to system...",FileName);
		exit(1);
	}
	
	pip->P[0] = AllocPES();
	scanned = fscanf(in,"%s %lf",S,&temp);
	if( (NewIPS(S))  && scanned == 2) {
		pip->P[0]->ExitSizeSpecified = 1;
		pip->P[0]->ExitSize = temp;
		pip->NZero = temp;
		pip->P[0]->Time = 0.0;
		pip->P[0]->TimeSpecified = 1;
	}
	else {
		printf("\n\nError!  Not finding InitPopSize on top line");
		exit(1);
	}
	
	/*  now enter a while loop to read in all the time points.  */
	i=0;
	
	while(!feof(in))  {  /*  while not yet at the end of the file */
		
		fscanf(in,"%s",S);  /*  read in the next string */
		
		/*  then decide what to do based on what that string is */
		if(NewR(S))  {
			fscanf(in,"%lf",&(pip->P[i]->Rate));
			pip->P[i]->RateSpecified = 1;
		}
		else if(NewEntry(S)) {
			fscanf(in,"%lf",&(pip->P[i]->EntrySize));
			pip->P[i]->EntrySizeSpecified = 1;
		}
		else if(NewExit(S)) {
			fscanf(in,"%lf",&(pip->P[i]->ExitSize));
			pip->P[i]->ExitSizeSpecified = 1;
		}
		else if(NewTime(S)) {
			fscanf(in,"%lf",&(pip->P[i]->Time));
			pip->P[i]->TimeSpecified = 1;
		}
		else if(NewSP(S)) {
			i++;
			pip->P[i] = AllocPES();
			pip->P[i]->IsSizePoint = 1;
		}
		else if(NewTP(S)) {
			i++;
			pip->P[i] = AllocPES();
			pip->P[i]->IsTimePoint = 1;
		}
	}
	
	/*  at the end of all that we set Num Points to i+1 and return the thing */
	pip->NumPoints = i+1;
	
	return(pip);
	
}



/*  this is a function that takes the pop-growth input  */
/*  information and converts it into the form that will */
/*  actually be used by other functions in the program */
popgrowth_container *ProcessPopGrowth(popgrowth_input_struct *PIP)
{
	int i;
	popgrowth_container *pc;
	double temp_sum;
	double step,t;
	int epoch;
	
	/*  some memory allocation and copying the number of different */
	/*  growth stages */
	pc = (popgrowth_container *)malloc(sizeof(popgrowth_container));
	pc->Num = PIP->NumPoints;
	pc->UsePopGrowth = 1;  /*  default it to use it if you've got it */
	pc->P = (popgrowth_struct *)calloc(pc->Num, sizeof(popgrowth_struct));
	
	/*  now cycle from present day (i=0) through all the other stages */
	/*  and enter and compute the information.   */
	/*  The way it gets done is this:  The rate, if present is put in there, */
	/*  if it's not there, then it is determined by the entry size at the next */
	/*  point.  If that is not there, then the exit size at the next time point  */
	/*  will be used.  If there is a rate and also an Entry sizepoint at a later time, then */
	/*  the rate gets overwritten.  Let's see how this takes form */
	
	for(i=0;i<pc->Num;i++)  {
		/*  set the time, if available */
		if(PIP->P[i]->TimeSpecified == 1 )  {
			pc->P[i].t = PIP->P[i]->Time;
			
			/*  then, if the time is specified, and */
			/*  if the entry size is specified (for i>0), then override the previous rate */
			/*  and print a message */
			if(PIP->P[i]->EntrySizeSpecified == 1  && (i > 0)  )  {
				pc->P[i-1].r = log(PIP->P[i]->EntrySize / PIP->P[i-1]->ExitSize)/
									(pc->P[i].t - pc->P[i-1].t)	;
									
				printf("\nStage %d: Rate of stage %d set to match entry size.",i,i-1);
			}
			/*  if the Entry size is not specified and there is no rate in the previous */
			/*  stage, then we are screwed.  so we exit. */
			else if((i > 0)   &&  PIP->P[i-1]->RateSpecified==0 )  {
				printf("\nStage %d: No entry size, and rate of stage %d not set.",i,i-1);
				printf("\nFATAL ERROR!!\n\n");
				exit(1);
			}
		}
		else {  /*  if the time is not available, then we must determine it.  This means */
				/*  that the current point must have an entry size, and we must have a rate */
				/*  in the previous stage */
			if(  (i > 0)  && PIP->P[i-1]->RateSpecified == 0 )  {
				printf("\nStage %d: No time, and rate of stage %d not set.",i,i-1);
				printf("\nFATAL ERROR!!\n\n");
				exit(1);
			}
			if(PIP->P[i]->EntrySizeSpecified == 0)  {					
				printf("\nStage %d: Time not set and no entry size.",i);
				printf("\nFATAL ERROR!!\n\n");
				exit(1);
			}
			else  {  /*  otherwise we set the time */
				pc->P[i].t = pc->P[i-1].t + log(PIP->P[i]->EntrySize / pc->P[i-1].Nplus)/
									(PIP->P[i-1]->Rate);
				
			}
			
		}
		
		
		/*  if the exit size is specified, then set it */
		if(PIP->P[i]->ExitSizeSpecified == 1)
			pc->P[i].Nplus = PIP->P[i]->ExitSize;
		/*  if not, then it will be the entry size, which we can now compute */
		/*  from the previous Nplus, the rate and the time */
		else if(i>0) {
			pc->P[i].Nplus = pc->P[i-1].Nplus * exp(pc->P[i-1].r *
					(pc->P[i].t - pc->P[i-1].t) );
		} 
		/*  if the rate is available, we set it too */
		if(PIP->P[i]->RateSpecified == 1)  {
			pc->P[i].r = PIP->P[i]->Rate;
		}
	}
	
	
	
	/*  OKAY!  Now that this is all set, we go back through and set the nu's and the sum of the */
	/*  nu's */
	temp_sum = 0.0;
	for(i=1;i<pc->Num;i++)  {
		if(pc->P[i-1].r != 0.0) 
			pc->P[i].nu = (1.0/(pc->P[i-1].r * pc->P[i-1].Nplus)) * (1.0 -
					exp(-pc->P[i-1].r * (pc->P[i].t - pc->P[i-1].t) ) );
		else 
			pc->P[i].nu = (pc->P[i].t - pc->P[i-1].t) / pc->P[i-1].Nplus;
			
		temp_sum += pc->P[i].nu;
		pc->P[i].sum_nu = temp_sum;
	}
	
	/*  and now, we want to fill out the ts and the ns for plotting the growth curves */
	/*  in GLUT.  We will go out till a time that is 2.0 times greater than the last time point */
	step = (pc->P[pc->Num-1].t * 2.0) / (double)NUM_POP_CURVE_POINTS;
	t = 0.0;
	pc->ts[0] = 0.0;
	pc->ns[0] = pc->P[0].Nplus;
	pc->MaxN = 0;
	epoch = 0;
	for(i=1;i<NUM_POP_CURVE_POINTS;i++)  {
		t += step;
		
		/*  set the time and the pop-size. */
		pc->ts[i] = t;
		pc->ns[i] = pc->P[epoch].Nplus * exp(pc->P[epoch].r * (t - pc->P[epoch].t) );
		
		/*  now, we have to get two points right on the next time point. */
		/*  so, if we have gone over the next time point, we reel back the */
		/*  current ts[i] to the entry N at the time point, and then set the */
		/*  next one to the Nplus at that time point.  */
		if(epoch < pc->Num-1)  {
			if(pc->ts[i] > pc->P[epoch+1].t)  {
				epoch++;
				t=pc->P[epoch].t;
				pc->ts[i] = t;
				pc->ns[i] = pc->P[epoch].Nplus * exp(pc->P[epoch].r * (t - pc->P[epoch].t) );
			
				if(pc->ns[i] > pc->MaxN)
					pc->MaxN = pc->ns[i];
				
				i++;
				pc->ts[i] = t;
				pc->ns[i] = pc->P[epoch].Nplus;
				
				if(pc->ns[i] > pc->MaxN)
					pc->MaxN = pc->ns[i];
			}
		}
		if(pc->ns[i] > pc->MaxN)
			pc->MaxN = pc->ns[i];
		
	}
	
	/*  and that ought to do it.  Return the pointer */
	return(pc);
}


/*  print out the PopGrowth settings */
void PrintPopGrowth(popgrowth_container *pc)
{
	int i;
	
	printf("\n\nPopulation growth settings as follows: (recall, t = 0");
	printf("\n\tis present.  Positive time is going back into the past...\n");
	
	for(i=0;i<pc->Num;i++)  {
		if(i<pc->Num-1)
			printf("\nStage %d:  Time = %.3f,  %.3f --> %.3f at rate %.12f.   nu = %.3f.  sum_nu = %.3f",i,
					pc->P[i].t, pc->P[i].Nplus, 
					pc->P[i].Nplus * exp(pc->P[i].r * (pc->P[i+1].t - pc->P[i].t) ),
					pc->P[i].r, pc->P[i].nu,
					pc->P[i].sum_nu);
		else
			printf("\nStage %d:  Time = %.3f,  %.3f --> ..... at rate %.12f.   nu = %.3f.  sum_nu = %.3f",i,
					pc->P[i].t, pc->P[i].Nplus, pc->P[i].r, pc->P[i].nu,
					pc->P[i].sum_nu);
	}
	printf("\n\n");
}


/*  returns time in the time-inhomogenous scale (but still in units of */
/*  twice the initial population size) given the population history and the */
/*  time in the homogenous scale (th). */
/*
	So, to be a little more explicit:  th is passed in and has units of
	Units generations (typically this will be 2N_{initial}).  It must be 
	converted into units of
	1 generation.  THEN all the computations can be made as far as
	the transformation, and the result is a value in units of 1 generation.
	At the very end this is divided by 2N_{initial} to return the
	result in units of 2N_{initial} generations
*/
double T_From_Th(popgrowth_container *pc, double th, double Units)
{
	double value;
	int i,istar;
	
	/*  convert the units of th to a single generation: */
	th *= Units;
	
	/*  our first task is to find i^*---the greatest i such that sum_nu < th */
	/*  this is not as fast as it could be---i.e. we could do a binary search for */
	/*  it, but with very few stages, it just won't make a difference. */
	istar = 0;
	for(i=0;i<pc->Num;i++)  {
		if(pc->P[i].sum_nu >= th / pc->P[0].Nplus)
			break;
		istar = i;
	}
	
	/*  now, istar should be defined appropriately, and we just do the transformation */
	if(pc->P[istar].r != 0)
		value = pc->P[istar].t - (1.0/pc->P[istar].r) * log(1.0 - (pc->P[istar].r * pc->P[istar].Nplus *
				( (th / pc->P[0].Nplus) - pc->P[istar].sum_nu ) ) );
	else 
		value = pc->P[istar].t + (pc->P[istar].Nplus *
				( (th / pc->P[0].Nplus) - pc->P[istar].sum_nu ) );
				
	/*  convert back to units of Units generations. */
	value /= Units;
	
	return(value);
}





/*  returns time in the time-homogeneous scale given the population */
/*  history and a time in the time-inhomogeneous scale.  Once again we */
/*  deal with the Units issue as above. */
double Th_From_T(popgrowth_container *pc, double t, double Units)
{
	double nu_star,value;
	int i,istar;
	
	/*  convert to units of single generations */
	t *= Units;
	
	/*  our first task is to find i^*---the greatest i such that t_i < t. */
	/*  this is not as fast as it could be---i.e. we could do a binary search for */
	/*  it, but with very few stages, it just won't make a difference. */
	istar = 0;
	for(i=0;i<pc->Num;i++)  {
		if(pc->P[i].t > t)
			break;
		istar = i;
	}
	
	/*  now, istar should be defined appropriately, and we just do the transformation */
	if(pc->P[istar].r != 0)
		nu_star = (1.0/(pc->P[istar].r * pc->P[istar].Nplus)) * (1.0 -
				(exp(-pc->P[istar].r * (t-pc->P[istar].t) )  ) );
	else 
		nu_star = (t - pc->P[istar].t) / pc->P[istar].Nplus;
		
	value = pc->P[0].Nplus * (nu_star + pc->P[istar].sum_nu);
				
	/*  convert back to units of Units generations */
	value /= Units;
	
	return(value);

}













