


/*  these are options for the type of simulation we are doing. */
#define NEW_HYBRIDS 0
#define PRITCHARD  	1
#define JABES		2
#define JABES_NO_PURES 3

GLOB int gAlphasConstrainedEqual;  /*  1 for yes 0 for no. */

#define JA_PURE 0
#define JA_ADMIXED 1


#define MAX_ALLELES 80
#define MAX_GN      80
#define MAX_LOCUS_NAME_LENGTH 80
#define MAX_CAT_NAME_LENGTH	  80
#define MAX_DATAFILE_NAME_LENGTH 100
#define MAX_NUM_SOURCES_PRITCH  20  /*  maximum number of components (sub-pops) for the Pritchard method */


#define CYCLE_alk(P) for(a=0;a<2;a++)  {for(l=0;l<P->L;l++)  {for(k=0;k<P->Kl[l];k++)  {
#define CYCLE_lka(P) for(l=0;l<P->L;l++)  {for(k=0;k<P->Kl[l];k++)  {for(a=0;a<2;a++)  {
#define CYCLE_ilc(P) for(i=0;i<P->M;i++)  {for(l=0;l<P->L;l++)  {for(c=0;c<2;c++)  {
#define END3CYCLE }}}

#define CYCLE_al(P) for(a=0;a<2;a++)  {for(l=0;l<P->L;l++)  {
#define DO_W_CYCLE(W) for(W[0]=0;W[0]<2;W[0]++) { for(W[1]=0;W[1]<2;W[1]++) {
#define CYCLE_TO_1(w) for(w=0;w<2;w++)  {
#define END2CYCLE  }}

#define CYCLE_i(P) for(i=0;i<P->M;i++)  {
#define CYCLE_g(P) for(g=0;g<P->Gn->v;g++)  {
#define CYCLE_l(P)   for(l=0;l<P->L;l++)  {
#define CYCLE_k(P) for(k=0;k<P->Kl[l];k++)  {
#define CYCLE_a for(a=0;a<2;a++)  {
#define CYCLE_c for(c=0;c<2;c++)  {
/*  here is one for cycling over the components in the Pritchard Methodm */
#define CYCLE_s(P) for(s=0;s<P->NumSources->v;s++)  { 
#define CYCLE_jv for(jv=0;jv<2;jv++)  {  /*  this one is for cycling over JA_PURE and JA_ADMIXED */


 
#define END1CYCLE  } 


/*  an enum for different types of loci */
enum locus_type
{
	UNSET,			/*  Not yet set  */
	CODOM, 			/*  codominant */
	AFLP,			/*  */
	HAS_NULL		/*  a locus that has a null allele */
};

/*  an enum for different types of priors */
enum prior_type
{
	UNIFORM,		/*  0 is for a uniform Dirichlet prior */
	JEFFREYS		/*  1 is for a Jeffreys Dirichlet prior */
};


/*  a struct to hold all the data, most of which won't get changed at all. */
typedef struct
{
	int *Kl;  	/*  the number of alleles in each locus */
	int L;		/*  the number of loci */
	int M;		/*  the number of individuals in the sample */
	
	int TypeOfSim;  /*  the declared type of simulation to be done at the beginning here */
	
	enum locus_type *LocTypes;  	/*  array of enums telling us what kinds of loci we have */

	char DataFileName[MAX_DATAFILE_NAME_LENGTH];
	char **LocNames;  /*  array of strings telling us the locus names */
	char ***AlleleNames;  	/*  array of array of strings giving the names of the alleles at all */
							/*  the loci. */
							
	ival ****Yobs;  	/*  the actual, OBSERVED data themselves, subscripted by [i][l][c]---i is the individual */
				/*  subscript, l is the locus subscript, c is the gene copy subscript (0 or 1, */
				/*  depending on if it is the first or second gene copy in the individual). */
				/*  values of Y which are >=0 will indicate different types of alleles, and */
				/*  == -1 will be our flag for a locus with missing data.   */
				/*  These OBSERVED values will be exactly what the Latent Data Y values are for */
				/*  codominant loci, but for other types of locus we shall have a difference */
				/*  FOR AFLP's Yobs[i][l][0] is 0 for  "-" individual and Yobs[i][l][0] is 1 for a "+" */
				/*  individual */
				
	ival *Gn;  /*  the number of genotype frequency classes  */
	
	char **CategoryNames;
	
	double ***G;	/*  the expected genotype frequencies of the Gn categories */
					/*  these two are included amongst the unkown in case I want */
					/*  to try sampling over them at some point.   */
					/*  G[g][0][0] is for 00, G[g][0][1] is for 01, */
					/*  G[g][1][0] is for 10, and G[g][1][1] is for 11 given the g-th category */
					/*  Note that the "/2" in equation 2 is already taken care of because we have */
					/*  G[g][0][1] = G[g][1][0].  ;   */
} hyb_data;



typedef struct 
{
	int SourceKnown;   /*  1 if the source is known for this individual, 0 otherwise */
	int Source;		   /*  takes a 0 or a 1 depending on which source the individual is from */
					   /*  These two imply the GtypFreqCatKnown == 1 for one of the pure categories. */
					   /*  But the index of that category is not determined until run time.  I will */
					   /*  need a function that uses SourceKnown and Source to assign values to */
					   /*  GtypFreqCatKnown and GtypFreqCat as appropriate. */
	
	int GtypFreqCatKnown;  /*  takes a 1 if the GtypFreqCategory is known, 0 otherwise */
	int GtypFreqCat;	/*  takes the value that Z for the individual should be set to */
	int AllelesContribute;  /*  1 if the alleles in the individual should be divvied up according */
							/*  to their W's when deriving the full conditional for the allele freqs. */
							/*  O if the individuals alleles should NOT contribute to the estimation  */
							/*  of frequencies */
	int AllelesToPrior;	/*  	1 if the alleles in this individual should contribute directly to the */
						/*  prior.  This should be 1 only if AllelesContribute==1 and SourceKnown==1. */
						/*  In that case, the alleles should contribute to the prior on allele frequency */
						/*  for the Source species.   */
	int ComputePofZ;		/*  1 if the PofZ should be computed for the individual.  This should only */
							/*  come into play if SourceKnown == 1 or GtypFreqCatKnown == 1 but we */
							/*  still want to ComputePofZ to see how well we would have been able to */
							/*  place the individual given our data and assumptions. */
	int ComputeLatentYsForKnownSource;   /*   if the individual is of known GtypFreqCat but has */
										 /*  AFLP's and so needs the latent  Y's to get summed  */
										 /*  over then this should be set to 1 and that calculation */
										 /*  will be carried out for the Pure GtypFreqCat to which the  */
										 /*  individual belongs, but only at the AFLP loci. */
	int ContributeToPi;	/*  Under the Gtyp Freq Categories model, if this is 1, then the individual's Z */
						/*  gets counted in computing the full conditional for Pi.  If 0, then the  */
						/*  individual's Z doesn't count toward computing the full cond for Pi. */

}  hyb_ind_flags;


/*  here is a simple struct for the latent variables for each individual that are */
/*  unique to Jonathan Pritchard's method. */
typedef struct 
{
	dval **Q;  /*  do this as a vector, so it extends to the multiple possible sources case easily. */
	ival **NumW;  /*  the number of genes, overall, in an individual, allocated to each component */
					/*  subscripted by [s] */
					
	dval **GtypProb;  /*  GtypProb[0] is given it is pure */
					  /*  GtypProb[1] is given it is admixed, and conditional on current */
					  /*  values of the alphas */
					  
	ival *V;  /*  the indicator of whether someting is pure (JA_PURE, 0) or admixed (JA_ADMIXED, 1) */
	dval **PofV;  /*  for storing the probabilities that an individual is pure or not. */
	
	
} Pritch_Indiv_Latent_struct;

typedef struct
{
	ival *NumSources;  /*  the number of components in the mixture */
	dval **Alpha;  /*  the alpha parameter---note that I will allow for non-symmetrical Dirichlet */
				  /*  priors on the Q variable */
	dval **MaxAlpha;  /*  max and min values that each component of alpha is allowed to take */
	dval **MinAlpha;
	
	dval **Sigma_SD;  /*   standard deviation of the normal proposal for elements of alpha.  This is a  */
					  /*  vector because we allow for different proposal widths for different components */

	dval **Xi;		/*  this is the mixing proportion between pures and admixed.  NOTE:  for Pi, we use the */
					/*  Pi that is already in the standard NewHybrids struct. */

	ival **Vcounts;  /*  the number of individuals currently in either the pure or admixed categories. */

} Pritch_Overall_Latent_struct;

/*  a struct to hold the latent variables that all apply to a single individual */
/*  in the sample */
typedef struct
{
	int idx;		/*  just stores the index of the individual.  This is useful for */
					/*  finding the Yobs for the individual in functions that just get */
					/*  a pointer to hyb_indiv passed to it.   */
					
	hyb_ind_flags *Flag;  /*  points to the flags for the individual that tell about its status. */
	
	ival ***Y;		/*  The unobserved data.  For codominant alleles, this will just be Yobs.  However, */
					/*  for other locus systems these values will change and be sampled over.  This is, */
					/*  however, the Y that all the probabilities of genotype will take place on. */
					/*  subscripted by [l][c] */
	
	ival ***W;	/*  an array similar in shape to that of Y in the data.  The elements indicate */
				/*  the species of origin of particular gene copies in the individuals */
				/*  subscripted by [l][c] where c = 0 or 1 */
	
	ival *Z;		/*  Z[i] points to an ival that indicates the genotype frequency */
					/*  class of the individual */
					
	dval **LogPofY;  /*  Log of the prob of the genotype given the current thetas and each of the different */
					/*  genotype frequency classes.  Subscripted by [g] */
					
	dval **PofZ;	/*  The probability that the individual belongs to each of the gtyp freq classes. */
					/*  subscripted by [g] */
					
	dval ****PofW;  /*  For each locus [l], the probability that the W's are 00, 01, 10, or 11 */
					/*  	subscripted by [l][w1][w2] where w1 \in {0,1} and w2 \in {0,1}.  Currently, for */
					/*  the Pritchard case, this is not used */
					
	double *****PofWandYlat;  /*  for each locus [l] and origin of markers [w1] or [w2] and type */
							/*  of AFLP marker 0, or 1, at copies one and two of the locus, this */
							/*  will store the probability of that. */
							/*  Subscripted by [l][w1][w2][y1][y2] */
							
	/*  a pointer to the Pritchard method extra latent variables for individuals (basically just */
	/*  Q)  */
	Pritch_Indiv_Latent_struct *PritInd;
	
					
} hyb_indiv;


/*  a struct to hold all the latent data and unknown parameters */
/*  and also to hold some space that will get used repeatedly */
typedef struct 
{
	int TypeOfSim;  /*  for the type of simulation: */
					/*  0 NewHybrid  (My genetics paper) */
					/*  1 Pritchard method */
					/*  2 My method from JABES paper.  These are defined above as constants */
	
	ival *Gn; 		/*  the number of genotype frequency categories */
	
	double ***G;	/*  the expected genotype frequencies of the Gn categories */
					/*  these two are included amongst the unkown in case I want */
					/*  to try sampling over them at some point.   */
					/*  G[g][0][0] is for 00, G[g][0][1] is for 01, */
					/*  G[g][1][0] is for 10, and G[g][1][1] is for 11 given the g-th category */
					/*  Note that the "/2" in equation 2 is already taken care of because we have */
					/*  G[g][0][1] = G[g][1][0].   */
					
	dval **Pi;  	/*  the mixing proportions.  Subscripted by [g] */
	
	dval ****Theta;  /*  The allele frequencies in the different populations */
					 /*  subscripted [a][l][k] */
	
	double ***r;  	 /*  this is used along the way---it counts the number of alleles */
					 /*  of the different types currently allocated to each of the two spp. */
					 /*  and then the prior will be added to it.  Subscripted [a][l][k]. */
					 
	double *s;  		 /*  used to count up the number of individuals in each geno. freq. cat. */
					 /*  similar to r. */
					 
	
	hyb_indiv **Ind;  /*  pointer to an array of individuals	 */
	
	sliding_trace_struct *CompleteDataLogLike;
	
	Pritch_Overall_Latent_struct *PritLat; 	
	
	dval **Locus_KB;   /*  an array of dvals subscripted by l.  Each carries the Kullback Leibler */
					   /*  divergence between the populations for each locus	 */
	
} hyb_latent;

typedef struct 
{
	dval ****Lambda;  	/*  the allele frequency priors */
	dval **Zeta;  		/*  the mixing proportion priors */
	
	dval **XiPrior;    /*  the priors on the mixing proportion of Pure vs Admixed in the JABES version */
	
} hyb_prior;

/*  A big monster struct to hold all info relative to a chain */
typedef struct
{
	hyb_data *Dat;
	hyb_latent *Lat;
	hyb_prior *Pri;
	
	int Seed1;
	int Seed2;

} hyb_chain;







/*  The function prototypes */
extern void Count_r(hyb_chain *C);
extern void Count_s(hyb_chain *C);
extern void UpdateTheta(hyb_chain *C);
extern void UpdatePi(hyb_chain *C);

extern double StoreSingleLogPofY(hyb_indiv *Ind, hyb_chain *C);
extern double PofY_at_a_locus(hyb_indiv *Ind,
		hyb_chain *C, int l, int g);
extern void UpdateWandZandY(hyb_chain *C);
extern int IntegerFromProbsRV(int Num,dval **P);
extern void DrawNewW_and_LatentY_indiv(hyb_indiv *Ind, hyb_chain *C);
extern void SingleSweep(hyb_chain *C);
extern void InitializeChain(hyb_chain *C);
extern hyb_indiv *CreateIndiv(hyb_data *D, int i);
extern hyb_chain *CreateLatentChain(hyb_data *D, hyb_prior *P);
extern hyb_prior * CreatePriors(hyb_data *D, 
		enum prior_type PiPriorType, enum prior_type ThetaPriorType);

extern void SetPiPriors(hyb_data *D, hyb_prior *P, enum prior_type PiPriorType);
extern void SetThetaPriors(hyb_data *D, hyb_prior *P, enum prior_type LambdaPriorType);

extern void IncrementValues(hyb_chain *C, int DoThetas);
extern void OutputHistograms(hyb_chain *C);
extern void ResetAllAveragesEtc(hyb_chain *C);
extern int YnotMissing(hyb_chain *C, int i, int l);

extern double KullbLeibAtLocus(hyb_chain *C, int l);
extern void KullbLeib(hyb_chain *C);


/*  data stuff */
extern hyb_data * GetData(char *FileName);
extern const char *LocType2str(enum locus_type loc);
extern const char *PriorType2str(enum prior_type loc);
extern void ReadDataLumpedFormat(int temp_int, int HasLocusNames, int ***TempY, 
				int TempM, int TempL, enum locus_type *TempLocTypes, int Divisor, FILE *in);
extern void ReadDataNonLumpedFormat(int temp_int, int HasLocusNames, int ***TempY, 
				int TempM, int TempL, enum locus_type *TempLocTypes, FILE *in);
extern void GetGtypFreqCats(hyb_data *D, char *FileName);

/*  Data output stuff */
void fprint_PofZ(hyb_chain *C);





/***************************************************************************
*
*		FUNCTIONS FOR PRITCHARD ET AL'S METHOD
*
***************************************************************************/
void PritchUpdateW(hyb_chain *C);
void PritchUpdateQ(hyb_chain *C);
Pritch_Indiv_Latent_struct *AllocPritchInd(int NumSources);
Pritch_Overall_Latent_struct *AllocPritchOverall(int NumSources);
void AllocatePritchardParts(hyb_chain *C);
void InitializePritchVars(hyb_chain *C);
void PritchUpdateAlphaRComp(hyb_chain *C, int r);
void PritchUpdateRandomCompOfAlpha(hyb_chain *C);
void IncrementPritchVars(hyb_chain *C, int DoThetas);
void PritchResetAllAveragesEtc(hyb_chain *C);
void PritchSingleSweep(hyb_chain *C);
void UpdateWbyBaum(hyb_chain *C);
void JABESComputePofZ(hyb_chain *C);
void JABESComputePofV(hyb_chain *C);
void JABESUpdateV(hyb_chain *C);
void JABEScount_Vcounts(hyb_chain *C);
void JABES_UpdateXi(hyb_chain *C);
void JABESUpdateAlphaUsingQ(hyb_chain *C, int r);
void JABESUpdateRandomCompOfAlpha(hyb_chain *C);
void JABES_UpdatePi(hyb_chain *C);
void JABES_UpdateZs(hyb_chain *C);
void JABES_UpdateWForPure(hyb_chain *C);
void JABES_StoreLogPofY_Pure(hyb_chain *C, int i, int a);
void UseLastFromNForPriorForZero(hyb_chain *C, int N);




















