#include <string.h>


#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "ECA_MemAlloc.h"
#include "MCTypesEtc.h"
#include "ECA_utilities.h"
#include "ECA_ReadData.h"
#include "GFMCMC.h"
#include "GFMCMC_DrawingUtilities.h"

#include "NewHybrids.h"



#define NH_LUMPED 0
#define NH_LINED 1
#define NH_NON_LUMPED 2


/*  prototypes for functions in this file: */
int NotInArray(int M, int *Array, int Num);
int ReturnSubscript(int *Array, int Length, int Compare);
void CheckLocType(char *inStr, enum locus_type *LocType, int i, int l);



int NotInArray(int M, int *Array, int Num) 
{
/*
	Checks to see if an integer is in an array of ints.
	Returns a 1 if M is not in the array.  0 otherwise.
*/
	int i;
	int NotInThere = 1;
	
	for(i=0;i<Num;i++)  {
		if(M == Array[i]) {
			NotInThere = 0;
			break;
		}
	}
	return(NotInThere);
}

/*  return the position of the first occurrence of Compare in Array of length Length */
/*  Returns a -1 if Compare is not in the array. */
int ReturnSubscript(int *Array, int Length, int Compare)
{
	int i;
	int temp=-1;
	
	for(i=0;i<Length;i++)  {
		if(Array[i] == Compare)	{
			temp = i;
			break;
		}
	}
	
	return(temp);
}

/*  interprets the locus input string and decides what locus type it is. */
/*  If the LocType is UNSET, it sets it to the appropriate type.  If the  */
/*  LocType is set to a different type, it barks a warning */
void CheckLocType(char *inStr, enum locus_type *LocType, int i, int l)
{	
	enum locus_type TempLT = CODOM;  	/*  start with the assumption that it is CODOMinant */
	
	if(strcmp(inStr,"+")==0 || strcmp(inStr,"-")==0) {
		TempLT = AFLP;
	}
	
	if(TempLT != *LocType)  {  /*  action of some sort will be necessary */
		if(*LocType == UNSET) {  /*  if it is unset, then we can set it */
			*LocType = TempLT;
		}
		else if(TempLT != *LocType){  /*  if it was already set to a different value, we have to bark an error message */
			 printf("\n\nError!  Locus %d, in Indiv %d, was %s.  But we just read",
			 		l,i,LocType2str(*LocType));
			 printf("a \'%s\' suggesting the locus is of type %s\n\n",inStr,LocType2str(TempLT));
		}
	}
}


hyb_data * GetData(char *FileName)
{
	int i,l,c,temp_int;
	int ***TempY;  /*  array of alleles carried at all loci by all individuals */
	int TempL,TempM, TempPolyL;	   /*  temporary number of loci and individuals */
	int *TempKl;  /*  temporary number of alleles at each locus */
	FILE *in;
	int **TempIdx;
	hyb_data *D;
	int locus_counter;
	char tempStr[500];  /*  a temporary variable to hold strings */
	char **TempLocNames;  /*  temporary, to hold locus names */
	enum locus_type *TempLocTypes;
	int HasLocusNames = 0;  /*  flag for whether it has locus names or not */
	int Digits, Divisor, matched;
	char Format[100];  /*  to record the format of the data */
	int TheFormat;
	
	
	
	/*  first get all the data into the Temp structures */
	in = erdOpenFileOrRetry(FileName, "r");
	
	
	if( (matched = fscanf(in,"NumIndivs %d  NumLoci %d Digits  %d  Format %s",
				&TempM,&TempL, &Digits, Format)) != 4) {
		printf("\n\nSorry, file %s does not begin in the format:",FileName);
		printf("\n\nNumIndivs XX\nNumLoci XX\nDigits XX\nFormat \"Format Type\" (i.e. Lumped of Lined)\n");
		printf("\n\nGo back and fix that and try again");
		printf("\n\nExiting to system\n\n");
		exit(1);
	}
	
	/*  set the value of the divisor */
	Divisor = (int)(pow(10.0f,Digits) + .001f);
	
	/*  set TheFormat */
	if( strcmp(Format,"Lumped")==0  || strcmp(Format,"LUMPED")==0  || strcmp(Format,"lumped")==0)
		TheFormat = NH_LUMPED;
	else if( strcmp(Format,"Lined")==0  || strcmp(Format,"LINED")==0  || strcmp(Format,"lined")==0 )
		TheFormat = NH_LINED;
	else if( strcmp(Format,"NonLumped")==0  || strcmp(Format,"NONLUMPED")==0  || strcmp(Format,"Nonlumped")==0  || strcmp(Format,"nonlumped")==0)
		TheFormat = NH_NON_LUMPED;
	else {
		printf("\n\nSorry! Data Format %s not recognized!\n\nExiting to system...",Format);
	}
	
	/*  allocate memory to the TempLocTypes, and initialize to UNSET */
	TempLocTypes = (enum locus_type *)ECA_MALLOC(TempM * sizeof(enum locus_type));
	for(l=0;l<TempL;l++)  {
		TempLocTypes[l] = UNSET;
	}
	
	/*  get the first string on the next line.  If it is "LocusNames" then */
	/*  collect the locus names */
	fscanf(in,"%s",tempStr);
	if(strcmp(tempStr,"LocusNames") == 0)  {
		HasLocusNames = 1;
		TempLocNames = (char **)ECA_MALLOC(TempL * sizeof(char *));
		for(l=0;l<TempL;l++)  {
			TempLocNames[l]  = (char *)ECA_MALLOC(MAX_LOCUS_NAME_LENGTH * sizeof(char));
			fscanf(in,"%s",TempLocNames[l]);
		}
	}
	else {  /*  if it is not LocusNames, then we just ate the number of the first individual */
			/*  in the data set, and so we send that to the variable temp_int */
		temp_int = atoi(tempStr);
	}
	
	
	/*  allocate all necessary memory to the TempY array: */
	TempY = (int ***)ECA_MALLOC(TempM * sizeof(int **));
	for(i=0;i<TempM;i++) {  /*  cycle over individuals, allocating necessary memory */
		TempY[i] = imatrix(0,TempL-1,0,1);
	}
	
	
	
	
	
	if(TheFormat == NH_LUMPED)  {
		ReadDataLumpedFormat(temp_int,HasLocusNames, TempY, 
					TempM, TempL, TempLocTypes, Divisor,  in);
	}
	if(TheFormat == NH_NON_LUMPED) {
		ReadDataNonLumpedFormat(temp_int,HasLocusNames, TempY, 
					TempM, TempL, TempLocTypes,  in);
	}
	
	
	
	
		
	/*  OK, now that this is all read into memory, we can start figuring out which of the loci */
	/*  are polymorphic and which locus indexes need shifting, etc.   */
	TempKl = ivector(0,TempL-1);
	TempIdx = imatrix(0,TempL-1,0,MAX_ALLELES);
	TempPolyL = 0;
	for(l=0;l<TempL;l++)  {
		TempKl[l] = 0;  /*  initialize to count the number of alleles */
		for(i=0;i<TempM;i++)  {
			for(c=0;c<2;c++)  {
				if(TempY[i][l][c] >= 0 ) {	/*  only count alleles that are not, for example, "-1" (as those are missing data.		 */
					if(NotInArray(TempY[i][l][c],TempIdx[l],TempKl[l]))  {
						TempIdx[l][TempKl[l]++] = TempY[i][l][c];
					}
				}
			}
		}
		if(TempKl[l] > 1)
			TempPolyL++;
	}
	
	/*  we now have the resources to copy all that stuff over to the hyb_data */
	/*  First allocate the memory */
	D = (hyb_data *)ECA_MALLOC(sizeof(hyb_data));
	sprintf(D->DataFileName,"%s",FileName);
	
	/*  default the type of simulation to NewHybrids for now */
	D->TypeOfSim = NEW_HYBRIDS;
	
	
	D->M = TempM;
	D->Kl = ivector(0,TempPolyL-1);
	D->Yobs = (ival ****)ECA_MALLOC(D->M * sizeof(ival ***));
	D->LocNames = (char **)ECA_MALLOC(TempPolyL * sizeof(char *));
	D->LocTypes = (enum locus_type *)ECA_MALLOC(TempPolyL * sizeof(enum locus_type));
	for(l=0;l<TempPolyL;l++)  {
		D->LocNames[l] = (char *)ECA_MALLOC(MAX_LOCUS_NAME_LENGTH * sizeof(char));
	}
	for(i=0;i<D->M;i++)  {
		D->Yobs[i] = (ival ***)ECA_MALLOC(TempPolyL * sizeof(ival **));
		for(l=0;l<TempPolyL;l++)  {
			D->Yobs[i][l] = IvalVector(0,1,0,0,0);
		}
	}
	/*  Then copy over the polymorphic loci with the different indexes */
	for(i=0;i<D->M;i++)  {
		locus_counter = 0;
		for(l=0;l<TempL;l++)  {
			/*  pick out only those loci that have more than 1 allele */
			if(TempKl[l] > 1) {
				/*  set the Y values.  If the loci are Codom we jumble the */
				/*  names of the alleles however we wish, but for AFLP's the  */
				/*  O's must remain 0's and so with the 1's */
				for(c=0;c<2;c++)  {
					if(TempLocTypes[l] == CODOM)  {
						if(TempY[i][l][c] == -1)  {  /*  if it is missing data, set Yobs to missing too. */
							D->Yobs[i][locus_counter][c]->v = -1;
						}
						else {
							D->Yobs[i][locus_counter][c]->v = ReturnSubscript(
										TempIdx[l],TempKl[l],TempY[i][l][c]);
						}
					}
					else if(TempLocTypes[l] == AFLP) 
						D->Yobs[i][locus_counter][c]->v = TempY[i][l][c];
					else {
						printf("\n\nABORT!  We have a locus that is neither CODOM or AFLP\n\n");
						exit(1);
					}
						
				}
				locus_counter++;
			}
		}
	}
	
	/*  and now we should be able to do the allele numbers and locus numbers, types, and names, too */
	D->L = TempPolyL;
	locus_counter = 0;
	for(l=0;l<TempL;l++)  {
		if(TempKl[l] > 1) {
			/*  set the LocusNames and types */
			D->LocTypes[locus_counter] = TempLocTypes[l];
			if(HasLocusNames) 		/*  if they have names, then copy them */
				sprintf(D->LocNames[locus_counter],"%s",TempLocNames[l]);
			else  /*  if not, then give them a number from 1 to the number of polymorphic loci */
				sprintf(D->LocNames[locus_counter],"%d",locus_counter+1);
			D->Kl[locus_counter] = TempKl[l];
			locus_counter++;
		}
	}
	
/*	
	out = fopen("ProcessedDataEcho.txt","w");
	fprintf(out,"%d  %d",D->M, D->L);
	for(i=0;i<D->M;i++)  {
		fprintf(out,"\n%d ",i+1);
		for(l=0;l<D->L;l++)  {
			fprintf(out,"  %d%d",D->Yobs[i][l][0]->v,D->Yobs[i][l][1]->v);
		}
	}
	fclose(out);
*/	
	
	
	
	/*  now, free all the temporary memory */
	for(i=0;i<TempM;i++)  {
		free_imatrix(TempY[i],0,TempL-1,0,1);
	}
	free(TempY);
	free_ivector(TempKl,0,TempL-1);
	free_imatrix(TempIdx,0,TempL-1,0,MAX_ALLELES);
	free(TempLocTypes);
	if(HasLocusNames) {
		for(l=0;l<TempL;l++)  {
			free(TempLocNames[l]);
		}
		free(TempLocNames);
	}
	
	return(D);
}


/*  a function to read the genotype frequency class data from a file named FileName. */
/*  it allocates memory in D for this information */
void GetGtypFreqCats(hyb_data *D, char *FileName)
{
	int g, w[2];
	FILE *in, *out;
	
	/*  We can get all the gtyp freq data now too. */
	in = erdOpenFileOrRetry(FileName,"r");
	
	
	D->Gn = AllocIval(0,0,0);
	fscanf(in,"%d",&(D->Gn->v));
	/*  allocate memory to G and to CategoryNames */
	D->G = (double ***)ECA_MALLOC(D->Gn->v * sizeof(double **));
	D->CategoryNames = (char **)ECA_CALLOC((size_t)D->Gn->v,sizeof(char *));
	
	for(g=0;g<D->Gn->v;g++)  {
		D->G[g] = dmatrix(0,1,0,1);  /*  allocate memory */
		D->CategoryNames[g] = (char *)ECA_CALLOC(MAX_CAT_NAME_LENGTH, sizeof(char));
		/*  read the name of the category first */
		fscanf(in,"%s",D->CategoryNames[g] );
		/*  read the gtyp freq quantities */
		DO_W_CYCLE(w)
			fscanf(in,"%lf",&(D->G[g][w[0]][w[1]]) );
		END2CYCLE
	}
	fclose(in);
	
	
	/*  now print that to make sure it is ok */
	out = fopen("EchoedGtypFreqCats.txt","w");
	for(g=0;g<D->Gn->v;g++)  {
		fprintf(out,"\n");
		DO_W_CYCLE(w)
			fprintf(out,"%.8f\t",D->G[g][w[0]][w[1]] );
		END2CYCLE
	}
	fclose(out);
	
}


/*  a function to read in the data when each individual's genotype is given */
/*  by a single number that must be split up. */
/*  temp_int is the first individual number if HasLocusNames == 0 */
/*  otherwise it just gets set to the individual number and then it */
/*  gets used throughout.  MISSING DATA IS DENOTED IN TEMPY BY -1 */
void ReadDataLumpedFormat(int temp_int, int HasLocusNames, int ***TempY, 
				int TempM, int TempL, enum locus_type *TempLocTypes, int Divisor, FILE *in)
{
	int i,l;
	char tempStr[100];
	FILE *out;
	
	
	for(i=0;i<TempM;i++) {  /*  cycle over individuals */
		if(HasLocusNames==1 || i > 0) {
			fscanf(in," %s",tempStr);  /*  eat the first number (the fish index) */
			temp_int = atoi(tempStr);
		}
		if(temp_int != i+1)
			printf("\nEating the wrong fish number!!\n\n");
		
		/*  now cycle over loci and fill them in.  Get the locus thing first as a string */
		/*  then, after recording what type of locus it is, record it in Y. */
		for(l=0;l<TempL;l++)  {
			fscanf(in,"%s",tempStr);
			
			/*  the following function checks AND sets the locus_type */
			CheckLocType(tempStr,&TempLocTypes[l],i,l);
			if(TempLocTypes[l]==CODOM) {
				temp_int = atoi(tempStr);
				TempY[i][l][0] = temp_int / Divisor;
				TempY[i][l][1] = temp_int % Divisor;
				
				/*  call the locus missing data if either of the TempY's is zero */
				if(TempY[i][l][0] == 0 || TempY[i][l][1]==0)  {
					TempY[i][l][0] = -1;
					TempY[i][l][1] = -1;  
				}
			}
			else if(TempLocTypes[l]==AFLP) {
				if(strcmp(tempStr,"+")==0)  {
					TempY[i][l][0] = 1;
					TempY[i][l][1] = 1;
				}
				else if(strcmp(tempStr,"-")==0) {
					TempY[i][l][0] = 0;
					TempY[i][l][1] = 0;
				}
				else 
					printf("\n\nError! Indiv %d, Locus %d.  AFLP but not + or -\n\n",i,l);
			}
			
		}
	}
	fclose(in);
	
	
	/*  print out the input to a file for comparison */
	out = fopen("EchoedGtypData.txt","w");
	for(i=0;i<TempM;i++) {  /*  cycle over individuals		 */
		fprintf(out,"\n%d",i+1);  /*  eat the first number (the fish index) */
	
		/*  now cycle over loci and fill them in */
		for(l=0;l<TempL;l++)  {
			fprintf(out,"  %d%d",TempY[i][l][0],TempY[i][l][1]);
			
		}
		
	}
	fclose(out);
	
}





/*  a function to read in the data when each individual's genotype is given */
/*  by a two numbers adjacent to each other on a line */
/*  temp_int is the first individual number if HasLocusNames == 0 */
/*  otherwise it just gets set to the individual number and then it */
/*  gets used throughout. */
void ReadDataNonLumpedFormat(int temp_int, int HasLocusNames, int ***TempY, 
				int TempM, int TempL, enum locus_type *TempLocTypes, FILE *in)
{
	int i,l;
	char tempStr[100];
	FILE *out;
	
	
	for(i=0;i<TempM;i++) {  /*  cycle over individuals */
		if(HasLocusNames==1 || i > 0) {
			fscanf(in," %s",tempStr);  /*  eat the first number (the fish index) */
			temp_int = atoi(tempStr);
		}
		if(temp_int != i+1)
			printf("\nEating the wrong fish number!!\n\n");
		
		/*  now cycle over loci and fill them in.  Get the locus thing first as a string */
		/*  then, after recording what type of locus it is, record it in Y. */
		for(l=0;l<TempL;l++)  {
			fscanf(in,"%s",tempStr);
			
			/*  the following function checks AND sets the locus_type */
			/*  if it is codominant, it has to read the other allele as well. */
			/*  if it is AFLP it does not have to read the other allele and just */
			/*  records the info and proceeds */
			CheckLocType(tempStr,&TempLocTypes[l],i,l);
			if(TempLocTypes[l]==CODOM) {
				temp_int = atoi(tempStr);
				TempY[i][l][0] = temp_int;
				
				/*  then also grab the next allele */
				fscanf(in,"%s",tempStr);
				temp_int = atoi(tempStr);
				TempY[i][l][1] = temp_int;
				
				/*  then check to see if they were missing data.  Bark an error and exit it */
				/*  only one of the two was missing */
				if(TempY[i][l][0] <= 0 && TempY[i][l][1] <= 0)  {
					TempY[i][l][0] = -1;
					TempY[i][l][1] = -1;
				}
				else if(  (TempY[i][l][0] > 0 && TempY[i][l][1] <= 0) )  {
					printf("\n\nError!  Gene Copy 0 Non-Missing and Gene Copy 1 Missing");
					printf("\nAt Locus %d on Individual %d\n\n", l+1,  i+1);
					printf("Exiting to system...\n\n");
					exit(1);
				}
				
				else if(TempY[i][l][0] <= 0 && TempY[i][l][1] > 0) {
					printf("\n\nError!  Gene Copy 1 Non-Missing and Gene Copy 0 Missing");
					printf("\nAt Locus %d on Individual %d\n\n", l+1,  i+1);
					printf("Exiting to system...\n\n");
					exit(1);
				}
						   
			}
			else if(TempLocTypes[l]==AFLP) {
				if(strcmp(tempStr,"+")==0)  {
					TempY[i][l][0] = 1;
					TempY[i][l][1] = 1;
				}
				else if(strcmp(tempStr,"-")==0) {
					TempY[i][l][0] = 0;
					TempY[i][l][1] = 0;
				}
				else 
					printf("\n\nError! Indiv %d, Locus %d.  AFLP but not + or -\n\n",i,l);
			}
			
		}
	}
	fclose(in);
	
	
	/*  print out the input to a file for comparison */
	out = fopen("EchoedGtypData.txt","w");
	for(i=0;i<TempM;i++) {  /*  cycle over individuals		 */
		fprintf(in,"\n%d",i+1);  /*  print the first number (the fish index) */
	
		/*  now cycle over loci and fill them in */
		for(l=0;l<TempL;l++)  {
			fprintf(in,"   %d %d",TempY[i][l][0],TempY[i][l][1]);
			
		}
		
	}
	fclose(out);
	
}




























