#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "ECA_MemAlloc.h"
#include "MCTypesEtc.h"
#include "MathStatRand.h"
#include "ranlib.h"
#include "GFMCMC.h"
#include "GFMCMC_DrawingUtilities.h"

#include "NewHybrids.h"

/*  updates all the variables in a single chain */
/*  I do this in the order of W and Z first, then theta and */
/*  pi, because it is easier to initialize Pi and Theta (and the Z's) */
/*  than initializing W's and Z's */
void SingleSweep(hyb_chain *C)
{
	/*  Do all the updates specific for each type of simulation */
	
	if(C->Lat->TypeOfSim == NEW_HYBRIDS) {
		UpdateWandZandY(C);
		UpdatePi(C);
	}	
	
	/*  here, this one gets done always */
	UpdateTheta(C);
	
	/*  compute the Kullback-Leibler divergence between the populations */
	KullbLeib(C);
	
}



/*  adds the current realization to the averages and standard errors and this histograms */
/*  of all the variables of interest. */
/*  If DoThetas = 0, then it won't increment the thetas.  This will be useful when  */
/*  we start doing multiple sampling locations */
void IncrementValues(hyb_chain *C, int DoThetas)
{
	int a,l,k,g,i;
	
	/*  increment the thetas */
	if(DoThetas)  {
		CYCLE_alk(C->Dat)
			IncrementDval(C->Lat->Theta[a][l][k]);
		END3CYCLE
	}
	
	/*  increment pi */
	CYCLE_g(C->Lat) 
		IncrementDval(C->Lat->Pi[g]);
	END1CYCLE
	
	/*  now cycle over the individuals and take care of PofZ  */
	/*  in the future I will take care of W, PofW, and the unobserved */
	/*  values for dominant marker types... */
	CYCLE_i(C->Dat)
		CYCLE_g(C->Lat)
			IncrementDval(C->Lat->Ind[i]->PofZ[g]);
		END1CYCLE
	END1CYCLE
	
	CYCLE_l(C->Dat)
		IncrementDval(C->Lat->Locus_KB[l]);
	END1CYCLE
	
}

/*  allocate memory to a hyb_indiv_latent struct, and set the pointers */
/*  for the allelic types to the observed data for codominant locus types. */
/*  It needs the i to set the pointers correctly to the i-th individual */
hyb_indiv *CreateIndiv(hyb_data *D, int i)
{
	int l,w1;
	int w[2],y[2];
	hyb_indiv *temp;
	
	/*  allocate to temp */
	temp = (hyb_indiv *)ECA_MALLOC(sizeof(hyb_indiv));
	
	
	/*  allocate pointers and/or memory to the observed data Y, and to W while we are at it */
	/*  and to PofW as well */
	temp->Y = (ival ***)ECA_MALLOC(D->L * sizeof(ival **));
	temp->W = (ival ***)ECA_MALLOC(D->L * sizeof(ival **));

/*  if(Ind->WsGetSampled==1)  { */
	temp->PofW = (dval ****)ECA_MALLOC(D->L * sizeof(dval ***));
	temp->PofWandYlat = (double *****)ECA_MALLOC(D->L * sizeof(double ****));
/*  } */
/*  else "If there are AFLP's in the data, allocate to temp->PofYlatGivenW" */
/*  this will hold the values for indivs from fixed sources for drawing new Y' */


	/*  set the index of the individual, too */
	temp->idx = i;
	
	CYCLE_l(D)  /*  cycle over the loci */
		temp->Y[l] = (ival **)ECA_MALLOC(2 * sizeof(ival *));
		temp->W[l] = IvalVector(0,1, 0,0,-1);
		
		temp->PofW[l] = (dval ***)ECA_MALLOC(2*sizeof(dval **));
		for(w1=0;w1<2;w1++)  {
			temp->PofW[l][w1] = DvalVector(0,1, 0.0,0.0,-1.0);
		}
		temp->PofWandYlat[l] = NULL;  /*  set it originally to NULL, which is appropriate for  */
								   /*  CODOM markers */
								   
		if(D->LocTypes[l] == AFLP)  {  /*  but if the locus is AFLP, then we must allocate space */
									   /*  to the thing.  This is going to look a little ugly. */
/*  if(Ind->WsGetSampled==1)  {			 */
			/*  For the joint probs of Y and W: */
			temp->PofWandYlat[l] = (double ****)ECA_MALLOC(2 * sizeof(double ***));
			CYCLE_TO_1(w[0])
				temp->PofWandYlat[l][w[0]] = (double ***)ECA_MALLOC(2 * sizeof(double **));
			CYCLE_TO_1(w[1])
				temp->PofWandYlat[l][w[0]][w[1]] = (double **)ECA_MALLOC(2 * sizeof(double *));
			CYCLE_TO_1(y[0])
				temp->PofWandYlat[l][w[0]][w[1]][y[0]] = (double *)ECA_CALLOC(2,sizeof(double));
			END3CYCLE	
/*  } */
/* else  "allocate to a smaller set of PofYsGivenW"   //} */
			/*  For the actual Y value at this locus */
			temp->Y[l] = IvalVector(0,1, 0, 0, 0);					   
			
		}
		
	END1CYCLE
	
	/*  Then cycle over the gtyp freq categories and allocate space for PofZ and LogPofY */
	temp->LogPofY = DvalVector(0,D->Gn->v-1, 0.0, 0.0, -1.0);
	temp->PofZ = DvalVector(0,D->Gn->v-1, 0.0, 1.0, .01);
	
	/*  and allocate space to Z too */
	temp->Z = AllocIval(0,0,0);
	
	
	/*  And, now for the exciting part, we point all the observed allele pointers to */
	/*  the values in the data struct for loci of type CODOM.   */
	/*  However, for the AFLP loci we don't do that, since these already point to memory */
	/*  that got allocated from them above.    AND!! IF THE LOCUS WAS NOT OBSERVED */
	/*  I.E. WAS MISSING!! THEN WE HAVE TO ALLOCATE MEMORY TO Y.  THEN WE NEED TO WRITE */
	/*  ANOTHER FUNCTION THAT IMPUTES THE VALUES OF Y IF Yobs == -1.  THIS IMPUTATION WILL */
	/*  BE BASED ON THE CURRENT VALUE OF z, I BELIEVE. */
	
	CYCLE_l(D)  /*  cycle over the loci */
		if(D->LocTypes[l] == CODOM) {
			for(w1=0;w1<2;w1++)  {
				temp->Y[l][w1] = D->Yobs[i][l][w1];
			}
		}
	END1CYCLE
	
	
	return(temp);
	
}




/*  Allocates all the necessary memory for the latent data associated with one chain */
/*  it uses the information in the data struct to determine what is necessary */
hyb_chain *CreateLatentChain(hyb_data *D, hyb_prior *P)
{	
	int a,l,i;
	hyb_chain *C;
	
	/*  allocate to C */
	C = (hyb_chain *)ECA_MALLOC(sizeof(hyb_chain));
	
	/*  first thing:  make the data pointer in C point to D */
	C->Dat = D;
	
	/*  then make the prior pointer point to the Priors */
	C->Pri = P;
	
	/*  Now allocate to the latent struct */
	C->Lat = (hyb_latent *)ECA_MALLOC(sizeof(hyb_latent));
	
	/*  make the Gn and G pointers in Lat point to the appropriate places in D */
	C->Lat->Gn = D->Gn;
	C->Lat->G  = D->G;
	
	/*  allocate to Pi array and the s-array */
	C->Lat->Pi = DvalVector(0,C->Lat->Gn->v-1, 0.0, 1.0, .01);
	C->Lat->s = (double *)ECA_CALLOC((size_t)C->Lat->Gn->v-1,sizeof(double));
	
	/*  by default, make the type of simulation NewHybrids */
	C->Lat->TypeOfSim = NEW_HYBRIDS;
	
	/*  allocate to Theta arrays and r->arrays */
	C->Lat->Theta = (dval ****)ECA_MALLOC(2*sizeof(dval ***));
	C->Lat->r = (double ***)ECA_MALLOC(2*sizeof(double **));
	CYCLE_a
		C->Lat->Theta[a] = (dval ***)ECA_MALLOC(C->Dat->L * sizeof(dval **));
		C->Lat->r[a] = (double **)ECA_MALLOC(C->Dat->L * sizeof(double *));
	END1CYCLE
	
	CYCLE_al(C->Dat)
		C->Lat->Theta[a][l] = DvalVector(0,C->Dat->Kl[l]-1, 0.0, 1.0, .01);
		C->Lat->r[a][l] = (double *)ECA_CALLOC((size_t)C->Dat->Kl[l]-1,sizeof(double));
		
	END2CYCLE
	
	/*  allocate to the array of pointers to individuals */
	C->Lat->Ind = (hyb_indiv **)ECA_MALLOC(C->Dat->M * sizeof(hyb_indiv));
	
	/*  then cycle over the individuals and allocate space to each one of them */
	CYCLE_i(C->Dat)
		C->Lat->Ind[i] = CreateIndiv(D,i);
	END1CYCLE
	
	/*  then allocate space for a sliding trace struct that holds up to 1,000 entries */
	C->Lat->CompleteDataLogLike = gfduAllocSlidingTrace(1000,1);
	
	/*  and allocate space for the Kullback-Leibler divergence of each locus */
	C->Lat->Locus_KB = DvalVector(0,C->Dat->L-1, 0.0, 0.0, 0.0);
	
	return(C);

}


/*  Create Jeffrey's Priors for the Pi and Theta structs */
/*  This returns a pointer to the priors and takes as */
/*  input the Data, D */
hyb_prior * CreatePriors(hyb_data *D, enum prior_type PiPriorType, enum prior_type ThetaPriorType)
{
	int l,a;
	hyb_prior *P;
	
	/*  first allocate the necessary memory */
	/*  To Lambda (prior parameters for Theta) */
	P = (hyb_prior *)ECA_MALLOC(sizeof(hyb_prior));
	P->Lambda  = (dval ****)ECA_MALLOC(2*sizeof(dval ***) );
	CYCLE_a
		P->Lambda[a] = (dval ***)ECA_MALLOC(D->L * sizeof(dval**));
		CYCLE_l(D)
			P->Lambda[a][l] = DvalVector(0,D->Kl[l]-1, 0.0,0.0,-1.0); 
		END1CYCLE
	END1CYCLE
	/*  to Zeta (prior parameters for Pi) */
	P->Zeta = DvalVector(0,D->Gn->v-1, 0.0,0.0,-1.0);
	
	/*  then set the values */
	SetPiPriors(D,P,PiPriorType);
	SetThetaPriors(D,P,ThetaPriorType);
	
	/*  may as well allocate and set values for the JABES related variable */
	P->XiPrior = DvalVector(0,1,0.0,1.0,.02);
	/*  and set them to Jeffreys right now too: */
	P->XiPrior[0]->v = .5;
	P->XiPrior[1]->v = .5;
	
	return(P);
}



/*  puts  Prior values in for the mixing proportions */
/*  assumes that memory has already been allocated to Zeta array */
void SetPiPriors(hyb_data *D, hyb_prior *P, enum prior_type PiPriorType)
{
	int g;
	
	CYCLE_g(D)
		if(PiPriorType == JEFFREYS)
			P->Zeta[g]->v = 1.0/(double)(D->Gn->v);
		else if(PiPriorType == UNIFORM)
			P->Zeta[g]->v = 1.0;
		else  /*  the default is Jeffreys */
			P->Zeta[g]->v = 1.0/(double)(D->Gn->v);
	END1CYCLE
}

/*  puts  Prior values in for the mixing proportions */
/*  assumes that memory has already been allocated to Lambda arrays */
void SetThetaPriors(hyb_data *D, hyb_prior *P, enum prior_type ThetaPriorType)
{
	int a,l,k;
	
	CYCLE_alk(D)
		if(ThetaPriorType == JEFFREYS)
			P->Lambda[a][l][k]->v = 1.0/(double)(D->Kl[l]);
		else if(ThetaPriorType == UNIFORM)
			P->Lambda[a][l][k]->v = 1.0;
		else  /*  default is Jeffreys prior */
			P->Lambda[a][l][k]->v = 1.0/(double)(D->Kl[l]);
	END3CYCLE
}


/*  Initialize Gn, Theta, Pi, (and Z) to "overdispersed" starting values */
/*  by simulating Theta and Pi from Dirichlet Dsns given by their priors */
/*  and then simulating Z's from the resulting Pi. */
/*  THIS SHOULD ALSO INITIALIZE Ind->Y TO RANDOM ALLELIC VALUES FOR LOCI THAT */
/*  ARE MISSING DATA. */
void InitializeChain(hyb_chain *C)
{
	int a,l,i,g,k;
	double tempTheta[MAX_ALLELES], tempPrior[MAX_ALLELES];
	double tempPi[MAX_GN], tempPiPrior[MAX_GN];
	
	
	/*  YO!  FIGURE OUT WHAT VALUE YOU WILL USE TO SET GN */
	/*  TO SOMETHING HERE. */
	
	
	/*  cycle over species and loci, initializing Theta */
	CYCLE_al(C->Dat) 
		/*  store the priors in the temporary array */
		CYCLE_k(C->Dat)
			tempPrior[k] = C->Pri->Lambda[a][l][k]->v;
		END1CYCLE
		/*  simulate the new value */
		DirichletRV(tempPrior,C->Dat->Kl[l],tempTheta);
		
		/*  copy it into the theta struct */
		CYCLE_k(C->Dat)
			C->Lat->Theta[a][l][k]->v = tempTheta[k];
		END1CYCLE
	END2CYCLE
	
	
	
	/*  then put the Pi Priors into temp storage and draw the initial Pi */
	CYCLE_g(C->Lat)
		tempPiPrior[g] = C->Pri->Zeta[g]->v;
	END1CYCLE
	DirichletRV(tempPiPrior,C->Lat->Gn->v,tempPi);
			
	/*  then copy those Pi values into the appropriate places */
	CYCLE_g(C->Lat)
		C->Lat->Pi[g]->v = tempPi[g];
	END1CYCLE
	
	/*  then simulate Z for each individual by drawing directly from Pi */
	CYCLE_i(C->Dat)
		C->Lat->Ind[i]->Z->v = IntFromProbsRV(tempPi,0,C->Lat->Gn->v);
		
		/*  then set the W's randomly given Z for each individual. */
		
		/*  Also, set the W values for the known source individuals who are */
		/*  not getting prior-ized.  Even for the AFLP ones. */
	END1CYCLE

}





/*  determines the value of r---the number of each type of allele */
/*  currently allocated to one or the other of the populations.  This */
/*  is currently written with codominant loci in mind */
/*  THIS WILL HAVE TO BE CHANGED FOR OTHER TYPES OF LOCUS */
void Count_r(hyb_chain *C)
{
	int i,l,k,a,c;  /*   subscripts for individuals (i),loci (l), alleles (k),  */
				  /*   species (a) or gene-copies (c) */
	
	/*  first, initialize r-values to 0's */
	CYCLE_alk(C->Dat)
		C->Lat->r[a][l][k] = 0.0;
	END3CYCLE
	
	/*  then cycle over the individuals and the gene copies within them */
	/*  and count up the alleles which are allocated according to W */
	CYCLE_i(C->Dat)
		CYCLE_l(C->Dat)
			
			/*  only count up the non-missing alleles.  Test here before looking at each allele */
			/*  (if one is missing, the other is too!) */
			if(YnotMissing(C,i,l) ) {
				CYCLE_c
					C->Lat->r[ C->Lat->Ind[i]->W[l][c]->v ]
							 [ l ]
							 [ C->Lat->Ind[i]->Y[l][c]->v ]  += 1.0;
				END1CYCLE
			}
	END2CYCLE
}



/*  Determines the value of s---the number of individuals currently allocated to */
/*  each of the different hybrid classes */
void Count_s(hyb_chain *C)
{
	int g,i;
	
	/*  first, initialize s values to 0.0's */
	CYCLE_g(C->Lat)
		C->Lat->s[g] = 0.0;
	END1CYCLE
	
	/*  then cycle over the individuals and add 1.0's to s according to their Z's */
	CYCLE_i(C->Dat)
		C->Lat->s[  C->Lat->Ind[i]->Z->v ] += 1.0;
	END1CYCLE

}


/*  a single function that updates all the theta values */
void UpdateTheta(hyb_chain *C)
{
	int a,l,k;
	double temp[MAX_ALLELES];  /*  a temporary array of doubles */
	
	/*  first count up the r's */
	Count_r(C);
	
	/*  then add the priors to those */
	CYCLE_alk(C->Dat)
		C->Lat->r[a][l][k] += C->Pri->Lambda[a][l][k]->v;
	END3CYCLE
	
	/*  Then cycle over pops and loci, simulating new values into the temporary vector */
	/*  and then copying them into the Theta structs */
	CYCLE_al(C->Dat)
		/*  simulate the new value */
		DirichletRV(C->Lat->r[a][l],C->Dat->Kl[l],temp);
		
		/*  copy it into the theta struct */
		CYCLE_k(C->Dat)
			C->Lat->Theta[a][l][k]->v = temp[k];
		END1CYCLE
	END2CYCLE
}


/*  a single function that updates all the pi values */
void UpdatePi(hyb_chain *C)
{
	int g;
	double temp[MAX_GN];  /*  a temporary array of doubles */
	
	/*  first count up the r's */
	Count_s(C);
	
	/*  then add the priors to those */
	CYCLE_g(C->Lat)
		C->Lat->s[g] += C->Pri->Zeta[g]->v;
	END1CYCLE
	
	/*  then simulate a new value  */
	DirichletRV(C->Lat->s,C->Lat->Gn->v,temp);
	
	/*  and copy it into the Pi struct */
	CYCLE_g(C->Lat)
		C->Lat->Pi[g]->v = temp[g];
	END1CYCLE
	
	
}


/*  for a single individual, this computes and stores the Log of the probabilities */
/*  of the genotypes of the individual for all the possible Gn gtyp freq categories */
/*  Along the way, it stores the Probability of the W's for that individual given */
/*  its current gtyp freq category allocation (i.e. given its current Z). */
/*  We pass in C so that we know the limits to loop over (i.e. the number of loci, etc). */
double StoreSingleLogPofY(hyb_indiv *Ind, hyb_chain *C)
{	
	int g,l;

	/*  cycle over gtyp freq classes */
/*  if(Ind->Flag->LogPofY==1 || C->Flag->DOM_Present == 1)  */
	CYCLE_g(C->Lat)
	
/*  if(Ind->Flag->LogPofY==1 || Ind->ZisFixed) */
		/*  initialize the LogPofY[g] to 0.0 to accumulate a sum over loci */
		Ind->LogPofY[g]->v = 0.0;
		
		/*  cycle over loci */
		CYCLE_l(C->Dat)
			Ind->LogPofY[g]->v += log(PofY_at_a_locus(Ind,C,l,g));
		END1CYCLE  /*  l */
		
	END1CYCLE /*  g */
/*  } */
	return(Ind->LogPofY[Ind->Z->v]->v);  /*  return the prob of its gtyp given the class it is  */
										 /*  allocated to. */
}


/*  computes the probability of Y at a locus l conditional on the individual */
/*  belonging to gtyp frequency class g.  Basically this implements */
/*  Equation 2 in the paper, summing, but summing over the four possible W's */
/*  C gives us a way to access the array G of expected proportions of the different genotypes */
/*  for the g-th genotype frequency class. */
/*  Crucially, this function also stores the full conditionals for W in PofW, based on the current */
/*  z of the individual. */
/*  For an AFLP locus, it stores the appropriate probabilities of the pairs (W,Y) because */
/*  the Y's are latent in the AFLP case.    */
double PofY_at_a_locus(hyb_indiv *Ind,
		hyb_chain *C, int l, int g)
{
	int i;
	double temp;	/*  initialize to accumulate a product */
	double normo=0.0;  /*  initialize to accumulate a sum */
	int w[2], y[2];
	
	
	/*  	DEALING WITH MISSING DATA: */
	/*  for now, we are not imputing, so if the data are missing, we */
	/*  just return 1.0, whether or not the locus is Codominant or AFLP */
	if(!YnotMissing(C,Ind->idx,l))
		return(1.0);
		
	
	
	
	
	/*  CODOM ROUTINE */
	if(C->Dat->LocTypes[l] == CODOM)  {
		DO_W_CYCLE(w)   /*  cycle over the different possible W's to marginalize them out */
			temp = 1.0;  /*  initialize to accumulate a product for each w[0],w[1] pair. */
			
			/*  get the allele frequency part in there first */
			for(i=0;i<2;i++)  {
				temp *= C->Lat->Theta[ w[i] ][l][ Ind->Y[l][i]->v ]->v;
			}
			
			/*  multiply the G part in there.  This depends on the gtyp freq class */
			/*  g that we are interested in */
			temp *= C->Lat->G[g][ w[0] ][ w[1] ];
			
			normo += temp;
					
			/*  now deal with storing the PofW part for the locus */
			if(g==Ind->Z->v)  {
				Ind->PofW[l][ w[0] ][ w[1] ]->v = temp;
			}
		END2CYCLE
		
		/*  now, if we just did the g for which the individual is currently allocated */
		/*  then we need to cycle over the w's one more time to normalize the PofW */
		if(g==Ind->Z->v)  {
			DO_W_CYCLE(w)
				Ind->PofW[l][ w[0] ][ w[1] ]->v /= normo;
			END2CYCLE
		}
	} 
	
	
	/*  AFLP ROUTINE:  For all g except the current one, we compute the probability of Y  */
	/*  given it is a "-" */
	/*  and then we either use that, if it really is a "-" or subtract it from 1.0 if */
	/*  it really is a "+".  For g that is the current one, we sum over all the possibilities */
	/*  so we can store them and use them to draw the W and the latent Y later.  Note that */
	/*  the "+" allele will always be allele 1 and the "-" allele will always be allele 0.  */
	if(C->Dat->LocTypes[l] == AFLP)  {

		if(g != Ind->Z->v)  {
			DO_W_CYCLE(w)   /*  cycle over the different possible W's to marginalize them out */
							/*  conditional on the whole individual being "-" for the locus */
				temp = 1.0;  /*  initialize to accumulate a product for each w[0],w[1] pair. */
				
				/*  get the allele frequency part in there first */
				for(i=0;i<2;i++)  {
					temp *= C->Lat->Theta[ w[i] ][l][ 0 ]->v;   /*  "0" subscripts the "-" allele */
				}
				
				/*  multiply the G part in there.  This depends on the gtyp freq class */
				/*  g that we are interested in */
				temp *= C->Lat->G[g][ w[0] ][ w[1] ];
				
				normo += temp;
			END2CYCLE
			
			/*  now, if the individual is really a "-", then normo is the probability that */
			/*  we want.  But if not, then we want to reset normo to 1-normo; */
			/*  Note that Yobs[i][l][0] (NOT Yobs[i][l][1]) is where the 0 or 1, */
			/*  signifying a "-" or a "+" resides. */
			if(C->Dat->Yobs[Ind->idx][l][0]->v == 1) 
				normo = 1.0 - normo; 
		}
		
		
		/*  if g is the the class this individual is currently allocated to, then we have to cycle */
		/*  over both W and possible latent y's, and store all that stuff. */
		else {
			
			DO_W_CYCLE(w)
				DO_W_CYCLE(y)
					
					/*  compute the probs of the w and y pairs.  if it is going to */
					/*  be zero, evaluate that first, to short-circuit multiplication by the */
					/*  allele freqs.  Basically if either of y[0] or y[1] are 1 when the  */
					/*  observed state is "-"  (coded as zero) then the prob is zero.  The same */
					/*  goes for y[0] and y[1] both being 0 when the observed state is "+" */
					if(   (  (y[0] + y[1] > 0) && (C->Dat->Yobs[Ind->idx][l][0]->v == 0 ) )  ||
						  (  (y[0] + y[1] == 0) && (C->Dat->Yobs[Ind->idx][l][0]->v == 1 ) )	)
					{
						temp = 0.0;
					}
					else {
						/*  first put the G part in there */
						temp = C->Lat->G[g][ w[0] ][ w[1] ];
						
						for(i=0;i<2;i++)  {
							temp *= C->Lat->Theta[ w[i] ][l][ y[i] ]->v;
						}
					}
					
					normo += temp;
					Ind->PofWandYlat[l][w[0]][w[1]][y[0]][y[1]] = temp;
				END2CYCLE
			END2CYCLE
			
			/*  finally, cycle one last time to normalize the things */
			DO_W_CYCLE(w)
				DO_W_CYCLE(y)
					Ind->PofWandYlat[l][w[0]][w[1]][y[0]][y[1]] /= normo;
				END2CYCLE
			END2CYCLE
			
		}
	}
	
	
	/*  that's it then! */
	return(normo);
}



/*  updates the W's and Z's for all the individuals, and the Latent Y's */
/*  for the AFLP loci.   */
void UpdateWandZandY(hyb_chain *C) 
{
	int i,g;
	double normo;
	double TempCompleteDataLogLike = 0.0;  /*  prepare to accumulate a sum */
	
	CYCLE_i(C->Dat)
		/*  compute the genotype probs (and the full conditionals for the w's) */
		/*  for each individual */
		TempCompleteDataLogLike += StoreSingleLogPofY(C->Lat->Ind[i], C);
		
		/*  then we must apply Bayes law to use PofY and Pi to compute  */
		/*  PofZ.  In the future I may want to do some fancy stuff to prevent */
		/*  underflow here (i.e. add in and later pull out a constant of log something before */
		/*  something before exponentiating LogPofY to become a probability).  */
		
		normo = 0.0;  /*  initialize to accumulate a sum of probabilities */
		
		/*  cycle over the genotype frequency classes */
		CYCLE_g(C->Lat)
			/*  first store them as unnormalized probabilities */
			C->Lat->Ind[i]->PofZ[g]->v = exp(C->Lat->Ind[i]->LogPofY[g]->v) * 
										 C->Lat->Pi[g]->v;
										 
			/*  and add that to the normalization factor */
			normo += C->Lat->Ind[i]->PofZ[g]->v;
		END1CYCLE
		
		/*  now cycle again over g and normalize PofZ */
		CYCLE_g(C->Lat)
			/*  divide each one by normo */
			C->Lat->Ind[i]->PofZ[g]->v /= normo;
		END1CYCLE
		
		/*  now we are set to draw new values for the W's and the Z's for each individual */
		/*  draw the W's first */
		DrawNewW_and_LatentY_indiv(C->Lat->Ind[i],C);
		
		/*  then draw the Z for the individual (you CANNOT do this before */
		/*  drawing the new W, without screwing things up!! */
		C->Lat->Ind[i]->Z->v = IntegerFromProbsRV(C->Lat->Gn->v,C->Lat->Ind[i]->PofZ);		
		
	END1CYCLE
	
	/*  now, at the end of all that, we update the CompleteDataLogLike Trace */
	gfduIncrementSlidingTrace(C->Lat->CompleteDataLogLike,TempCompleteDataLogLike);
}



/*  returns a discrete random integer between 0 and Num-1  */
/*  based on probabilities stored in an array of pointers to dvals. */
/*  This assumes that the probabilities are subscripted 0,...,Num-1 */
/*  Does not check to make sure that the probabilities */
/*  sum to one.  Requires function ranf() from ranlib or elsewhere */
/*  to generate a Uniform double on (0,1). */
/*  This should go into the MathStatRand library, eventually  */
int IntegerFromProbsRV(int Num,dval **P)
{
	int i;
	double cum=0.0;
	double rando = (double)ranf();
	
	for(i=0;i<Num;i++)  {
		cum += P[i]->v;
		if(cum > rando)
			break;
	}
	if(i==Num && cum<1.0)  {
		fprintf(stderr,"\n\nCumulative Prob less than 1.0 in IntegerFromProbsRV\n\n");
		fprintf(stderr,"Exiting to system...\n\n");
		exit(1);
	}
	
	return(i);
}




/*  cycle over the all the loci in an individual and  */
/*  update the W values at each locus.  This requires that */
/*  PofW be computed previously (usually by a call to  */
/*  JointPofWandY_at_a_locus) and that the Z for the individual */
/*  has not been changed since then.   */
void DrawNewW_and_LatentY_indiv(hyb_indiv *Ind, hyb_chain *C)
{
	int l,w[2],y[2];
	double cum,rando;
	
	CYCLE_l(C->Dat)
	
		/*  ONLY DRAW A NEW W and LATENT Y if THE LOCUS IS NOT MISSING!!! */
		if(YnotMissing(C,Ind->idx,l)) {
			cum = 0.0;  /*  set to zero to accumulate a sum */
			rando = (double)ranf();
	/* if(Ind->SampleWs==1)  {		 */
			/*  CODOM ROUTINE: */
			if(C->Dat->LocTypes[l] == CODOM)  {
				DO_W_CYCLE(w) 
					cum += Ind->PofW[l][ w[0] ][ w[1] ]->v;
					if(cum>rando) {
						/*  set the two values of Ind->W accordingly */
						Ind->W[l][0]->v = w[0];
						Ind->W[l][1]->v = w[1];
						/*  then set the little w's each to 2 so that we exit the two for loops */
						w[0] = 2;
						w[1] = 2;
					}
				END2CYCLE
			}
			
			/*  AFLP ROUTINE */
			if(C->Dat->LocTypes[l] == AFLP)  {
				DO_W_CYCLE(w)
		 			DO_W_CYCLE(y)
						cum += Ind->PofWandYlat[l][ w[0] ][ w[1] ][ y[0] ][ y[1] ];
						if(cum>rando) {
							/*  set the two values of Ind->W and of Ind->Y accordingly */
							Ind->W[l][0]->v = w[0];
							Ind->W[l][1]->v = w[1];
							Ind->Y[l][0]->v = y[0];
							Ind->Y[l][1]->v = y[1];
							/*  then set the little w's and y's all to 2 so that we exit the two for loops */
							w[0] = 2;
							w[1] = 2;
							y[0] = 2;
							y[1] = 2;
						}
					END2CYCLE
				END2CYCLE
			}
	/*  }   // closes if(Ind->SampleWs==1) */

	/*  else if(C->Dat->LocTypes[l] == AFLP)  "Sample The Ys but not Ws at This AFLP Locus"  } */
		}
	
	END1CYCLE
}



/*  Writes out the histograms to a series of files */
void OutputHistograms(hyb_chain *C)
{
	int l,k,a,g;
	FILE *out;
	
	/*  deal first with the thetas: */
	fprint_HistTopRow(C->Lat->Theta[0][0][0]->Hist, "aa-Theta.hist");
	out = fopen("aa-Theta.hist","a");
	
	CYCLE_lka(C->Dat)
		fprintf(out,"\nLoc%d_Alle%d_Pop%d  ",l,k,a);
		fprint_HistLineProp(out,C->Lat->Theta[a][l][k]->Hist);
	END3CYCLE
	fclose(out);
	
	/*  deal next with the Pi's */
	fprint_HistTopRow(C->Lat->Theta[0][0][0]->Hist, "aa-Pi.hist");
	out = fopen("aa-Pi.hist","a");
	
	CYCLE_g(C->Lat) 
		fprintf(out,"\n%.5f/%.5f/%.5f  ",C->Lat->G[g][0][0],
									  C->Lat->G[g][0][1] + C->Lat->G[g][1][0],
									  C->Lat->G[g][1][1]);
		fprint_HistLineProp(out,C->Lat->Pi[g]->Hist);
	END1CYCLE
	fclose(out);
}






/*  converts a locus_type enum to a string */
const char *LocType2str(enum locus_type loc)
{

	switch (loc) {
	case CODOM: return "CODOMINANT";
	case AFLP: return "AFLP";
	case UNSET: return "UNSET";
	case HAS_NULL: return "HAS_NULL";
	default: return "!!Error-Unknown Locus Type!!";
	}
}


/*  converts a prior_type enum to a string */
const char *PriorType2str(enum prior_type loc)
{

	switch (loc) {
		case UNIFORM: return "UNIFORM";
		case JEFFREYS: return "JEFFREYS";
		default: return "!!Error-Unknown Prior Type!!";
	}
}



/*  Reset all the averages and things for all of the variables in a chain */
/*  Currently it only deals with the few variables I have been keeping track of */
void ResetAllAveragesEtc(hyb_chain *C)
{
	int a,l,k,g,i;
	/*  first do all the Thetas */
	CYCLE_alk(C->Dat)
		InitDvalSummaryToZero(C->Lat->Theta[a][l][k]);
	END3CYCLE
	
	/*  then do the Pi */
	CYCLE_g(C->Dat) 
		InitDvalSummaryToZero(C->Lat->Pi[g]);
	END1CYCLE

	/*  now cycle over the individuals and take care of PofZ  */
	/*  in the future I will take care of W, PofW, and the unobserved */
	/*  values for dominant marker types... */
	CYCLE_i(C->Dat)
		CYCLE_g(C->Lat)
			InitDvalSummaryToZero(C->Lat->Ind[i]->PofZ[g]);
		END1CYCLE
	END1CYCLE
	
	/*  then initialize the sliding trace structs, too */
	gfduInitSlidingTraceToZero(C->Lat->CompleteDataLogLike);
	
	/*  reset the averages of the kullback-leibler stuff too */
	CYCLE_l(C->Dat)
		InitDvalSummaryToZero(C->Lat->Locus_KB[l]);
	END1CYCLE

}


/*  returns 1 if locus l in individual i is not missing */
/*  and  0 if it is missing */
int YnotMissing(hyb_chain *C, int i, int l)
{
	int temp = 1;
	
	if(C->Dat->Yobs[i][l][0]->v == -1 ||  C->Dat->Yobs[i][l][1]->v == -1)  {
		temp = 0;
	}
	
	return(temp);
}



/*  computes the current value of the Kullback-Leibler divergence at a locus */
double KullbLeibAtLocus(hyb_chain *C, int l)
{
	int a,k;
	double temp = 0.0;  /*  initialize to accumulate a sum at this locus */
	
		
	CYCLE_a
		CYCLE_k(C->Dat)
			temp += C->Lat->Theta[a][l][k]->v * log(C->Lat->Theta[a][l][k]->v/
				C->Lat->Theta[(a+1)%2][l][k]->v);  /*  note the little (a+1)%2 trick */
												   /*  to make the denominator the freq */
												   /*  in the other population */
		END1CYCLE
	END1CYCLE
	
	
	/*  now return temp */
	return(temp);
	
	
}


/*  computes all the current Kullback-Leibler divergences at all the loci */
void KullbLeib(hyb_chain *C)
{
	int l;
	
	CYCLE_l(C->Dat)
		C->Lat->Locus_KB[l]->v = KullbLeibAtLocus(C,l);
	END1CYCLE
} 







/*  a silly, expedient function for the scottish cats data set:  starting from */
/*  individual N (numbered N, not its index!!), till the end of the data set, use those individuals to form the */
/*  prior for allele frequencies (along with whatever existing prior had been there), */
/*   and then set the number of individuals to N, so that */
/*  those final ones get dropped from the dataset. */
void UseLastFromNForPriorForZero(hyb_chain *C, int N)
{
	int i,l,c;
	
	for(i=N-1; i<C->Dat->M ; i++)  {
		CYCLE_l(C->Dat)
			if( YnotMissing(C,i,l) )  {
				for(c=0;c<2;c++)  {
					C->Pri->Lambda[0][l][  C->Dat->Yobs[i][l][c]->v ]->v += 1.0;
				}
			}
		END1CYCLE
	}
	
	/*  then, at the end of this we set the number of individuals to N - 1 */
	C->Dat->M = N-1;
		
}








