#define UN_EXTERN



#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "ranlib.h"

#include "MathStatRand.h"
#include "MCTypesEtc.h"

#include "GFMCMC.h"
#include "ECA_MemAlloc.h"


// some global variables
ival *gInd, *gMH;
int gN;
double gp;


// prototypes
void gfbsDrawProbsSkeleton(void);
void gfbsDrawIndBinom(void);
void gfbsDrawConsoleWindow(void);
void gfbsDrawRectHist(int_hist_struct *H, double Total, int Current);
void gfbsMH_Update(ival *L);
void gfbsDrawMHBinom(void);






void gfmUserDefd_ResetAllAverages(void)
{
	InitIvalToZero(gInd);
	InitIvalSummaryToZero(gMH);
		
	gNumSweepsAfterBurnIn = 0;
	
}

void gfmUserDefd_InitializeChain(void)
{
	InitIvalToZero(gInd);
	InitIvalToZero(gMH);
	
	gMH->v = 0;

	gNumSweeps = 0;
	
}

void gfmUserDefd_OneStepByGlobals(void)
{
		gInd->v = ignbin(gN,(float)gp);
		IncrementIval(gInd);
		
		gfbsMH_Update(gMH);
		IncrementIval(gMH);	
}

void gfmUserDefd_DefineWindows(void)
{
	int tempX[2];
	double tempP[2];
	GLfloat MaxProb;
	
	// determine the extent of this binomial distributions
	tempX[0] = (int)(gN * gp);
	tempX[1] = gN - tempX[0];
	tempP[0] = gp;
	tempP[1] = 1.0 - gp;
	MaxProb = (GLfloat)exp(LogMultinomialPMF(gN, tempP, tempX, 2));
	
	
	// name the views files 
	gsOUTPUT_FILE_PREFIX("Pedagogy");
	
	gsCONSOLE("\"Info\" Window");
		gsDRAW_FUNC(gfbsDrawConsoleWindow);
		
	gsNEW_WINDOW(1,	"Independent Realizations");
		gsDRAW_FUNC(gfbsDrawIndBinom);
		gsXLO(-.20 * gN);
		gsXHI(1.1 * gN);
		gsYHI(1.5 * MaxProb);
		gsYLO(-.25 * MaxProb);
		gsAXES_POS_QUADRANT;
		gsCOLOR_SCHEME(DEEP_BLUE);
		
	gsNEW_WINDOW(2,	"Markov Chain Realizations");
		gsDRAW_FUNC(gfbsDrawMHBinom);
		gsXLO(-.20 * gN);
		gsXHI(1.1 * gN);
		gsYHI(1.5 * MaxProb);
		gsYLO(-.25 * MaxProb);
		gsAXES_POS_QUADRANT;
		gsCOLOR_SCHEME(FISHER_PRICE);
		
}



void gfmUserDefd_DefineMenus(void)    // to define menus for the right click of the mouse
{
	return;  // no special right click menus for now
}


void gfmUserDefd_LastWords(void)
{
	return;  // no special last words here
}



// given gN and gp, this proposes an update via a simple
// random walk proposal to a new state and it accepts it 
// according to the MH ratio.  
void gfbsMH_Update(ival *L)
{
	double qpart, MH, topProb, bottomProb;
	int Lprime;
	
	int tempX[2];
	double tempP[2];
	
	
	// make the proposal
	qpart = 1.0;
	if(ranf()<.5)
		Lprime = L->v+1;
	else
		Lprime = L->v-1;
		
	// deal with the special cases on the end
	if(L->v == 0) {
		Lprime = 1;
		qpart = .5;
	}
	if(L->v == gN) {
		Lprime = gN - 1;
		qpart = .5;
	}
	
	if(Lprime == 0 || Lprime == gN)
		qpart = 2;
	
	// compute the bottom and top limit dsn probs:
	tempX[0] = L->v;
	tempX[1] = gN - tempX[0];
	tempP[0] = gp;
	tempP[1] = 1.0 - gp;
	bottomProb = LogMultinomialPMF(gN, tempP, tempX, 2);
	
	tempX[0] = Lprime;
	tempX[1] = gN - tempX[0];
	tempP[0] = gp;
	tempP[1] = 1.0 - gp;
	topProb = LogMultinomialPMF(gN, tempP, tempX, 2);
	
	// compute the MH ratio
	MH = log(qpart) +  topProb - bottomProb;
	
	// then accept if appropriate
	if(log(ranf())<MH)
		L->v = Lprime;
	
}




void gfbsDrawProbsSkeleton(void)
{
	int i;
	ColorScheme3f *CS = gWindowsSettings[glutGetWindow()]->ColorScheme;
	GLfloat Top,Left,Right;
	int tempX[2];
	double tempP[2];
	
	
	// set the color to the first series color
	glColor3fv(CS->Series[0]);
	
	// cycle over all the probabilities
	for(i=0;i<=gN;i++)  {
		// compute the binomial probability
		tempX[0] = i;
		tempX[1] = gN - tempX[0];
		tempP[0] = gp;
		tempP[1] = 1.0 - gp;
		Top = (GLfloat)exp(LogMultinomialPMF(gN, tempP, tempX, 2));
		Left = (GLfloat)i;
		Right = Left + 1.0f;
		glBegin(GL_LINE_LOOP);
			glVertex2f(Left,0.0f);
			glVertex2f(Left,Top);
			glVertex2f(Right,Top);
			glVertex2f(Right,0.0f);
		glEnd();
	}
	
	
}

void gfbsDrawIndBinom(void)
{	
	
	gfbsDrawRectHist(gInd->Hist, gInd->Hist->dTot, gInd->v);
	
	gfbsDrawProbsSkeleton();
	
	// then draw the axes
	gfmDrawXYAxes();
}

void gfbsDrawMHBinom(void)
{	
	
	gfbsDrawRectHist(gMH->Hist, gMH->Hist->dTot, gMH->v);
	
	gfbsDrawProbsSkeleton();
	
	// then draw the axes
	gfmDrawXYAxes();
}


// draw a histogram out of rectangles.  The height of the rectangles will be divided by 
// Total.
void gfbsDrawRectHist(int_hist_struct *H, double Total, int Current)
{
	int i;
	ColorScheme3f *CS = gWindowsSettings[glutGetWindow()]->ColorScheme;
	GLfloat Top,Left,Right, Bottom;
	
	int tempX[2];
	double tempP[2];
	GLfloat MaxProb;
	
	// determine the extent of this binomial distributions
	tempX[0] = (int)(gN * gp);
	tempX[1] = gN - tempX[0];
	tempP[0] = gp;
	tempP[1] = 1.0 - gp;
	MaxProb = (GLfloat)exp(LogMultinomialPMF(gN, tempP, tempX, 2));
	
	// set the color to the second series color
	glColor3fv(CS->Series[1]);
	
	// cycle over all the bins
	for(i=0;i<H->NumBins;i++)  {
		// deterine the proportions of the rectangles
		Top = (GLfloat)( (double)H->H[i] / Total) ;
		Left = (GLfloat)i;
		Right = Left + 1.0f;
		glRectf(Left,0.0f,Right,Top);
	}
	
	// now draw the little rectangle beneath the current position
	Top = -.03f * MaxProb;
	Bottom = -.12f * MaxProb;
	Left = (GLfloat)Current;
	Right = Left + 1;
	glRectf(Left,Bottom,Right,Top);
	
	
}

void gfbsDrawConsoleWindow(void)
{
	char temp[250];
	int CW = glutGetWindow();	
	ColorScheme3f *CS = gWindowsSettings[CW]->ColorScheme;

	
	
	glColor3fv(CS->Text);
	gfmRenderBitmapString(-.1f,1.0f,GLUT_BITMAP_HELVETICA_12, "Program: MCMC Pedagogy");
	gfmRenderBitmapString(-.1f,.9f,GLUT_BITMAP_HELVETICA_12, "Author: E.C. Anderson");
	
	sprintf(temp,"N: %d", gN);
	gfmRenderBitmapString(-.1f,.8f,GLUT_BITMAP_HELVETICA_12, temp);
	
	sprintf(temp,"p: %f",gp);
	gfmRenderBitmapString(-.1f,.7f,GLUT_BITMAP_HELVETICA_12, temp);
	
	sprintf(temp,"Num of Iterations: %d",gNumSweeps);
	gfmRenderBitmapString(-.1f,.3f,GLUT_BITMAP_HELVETICA_12, temp);
	
	if(gGo) 
		sprintf(temp,"RUNNING!");
	else 
		sprintf(temp,"STOPPED!");
		
	gfmRenderBitmapString(-.1f,.2f,GLUT_BITMAP_HELVETICA_12, temp);
	
}

int main(int argcp, char **argv)
{	
	
	// ask the user for n
	printf("\n\nWe will be simulating binomial random variables.");
	printf("\n\nInput N and p\n\n->");
	scanf("%d %lf",&gN,&gp);
	
	
	gInd = AllocIval(0,gN, 1);
	gMH = AllocIval(0,gN, 1);
	
	gMH->v = ignbin(gN,(float)gp);  // just set it to this for now.
	
	// here must call glutInit from within main, with the command line option
	// pointer, and then call gfmInitGFM, and that does it.  
	glutInit(&argcp, argv);
	gfmInitGFM();

	return(1);
	
}

